import numpy as np
from gsoup.viewer import gviewer
from gsoup import (
    structures,
    pixels_in_world_space,
    to_hom,
    ray_ray_intersection,
    swap_columns,
    reconstruct_pointcloud,
    load_image,
)

"""
a few example functions for using the viewer
"""

#### globals
ui_float = 0.0
poses = None
meshes_v = None
meshes_f = None
gv_scalar, gv_color, gv_vector, gf_scalar, gf_color = None, None, None, None, None
pcs_v = None
#### globals


def pcs_slider_callback():
    global ui_float, pcs_v, gv_scalar, gv_color
    changed, ui_float = gviewer.psim.SliderFloat(
        "step", ui_float, v_min=0, v_max=len(pcs_v)
    )
    if changed:
        if int(ui_float) >= len(pcs_v):
            ui_float = len(pcs_v) - 1
        if gv_scalar[0] is not None:
            gviewer.register_pointcloud(
                "pc", pcs_v[int(ui_float)], s=gv_scalar[int(ui_float)], radius=0.0006
            )
        elif gv_color[0] is not None:
            gviewer.register_pointcloud(
                "pc", pcs_v[int(ui_float)], c=gv_color[int(ui_float)], radius=0.0006
            )
        else:
            gviewer.register_pointcloud("pc", pcs_v[int(ui_float)], radius=0.0006)


def meshes_slider_callback():
    global ui_float, meshes_v, meshes_f, gv_scalar, gv_color, gv_vector, gf_scalar, gf_color
    changed, ui_float = gviewer.psim.SliderFloat(
        "step", ui_float, v_min=0, v_max=len(meshes_v)
    )
    if changed:
        if int(ui_float) >= len(meshes_v):
            ui_float = len(meshes_v) - 1
        float1, float2, float3, float4, float5 = (
            ui_float,
            ui_float,
            ui_float,
            ui_float,
            ui_float,
        )
        if gv_scalar[0] is None:
            float1 = 0
        if gv_color[0] is None:
            float2 = 0
        if gv_vector[0] is None:
            float3 = 0
        if gf_scalar[0] is None:
            float4 = 0
        if gf_color[0] is None:
            float5 = 0
        gviewer.register_mesh(
            "mesh",
            meshes_v[int(ui_float)],
            meshes_f[int(ui_float)],
            s_vertices=gv_scalar[int(float1)],
            c_vertices=gv_color[int(float2)],
            v_vertices=gv_vector[int(float3)],
            s_faces=gf_scalar[int(float4)],
            c_faces=gf_color[int(float5)],
            edge_width=1.0,
        )


def poses_slider_callback():
    global ui_float, poses
    edge_rad = 0.0005
    changed, ui_float = gviewer.psim.SliderFloat(
        "step", ui_float, v_min=0, v_max=len(poses)
    )
    if changed and poses is not None:
        if int(ui_float) >= len(poses):
            ui_float = len(poses) - 1
        v_tot, e_tot, c_tot = gviewer.register_camera(
            "poses", poses[int(ui_float)], edge_rad, group=True
        )


def pcs_slide_view(v, s_vertices=None, c_vertices=None):
    """
    given some point cloud tXVx3
    show the point cloud as it changes through time t and allow scrolling through using a slider.
    """
    global pcs_v, gv_scalar, gv_color
    gviewer.init()
    gviewer.ps.set_up_dir("z_up")
    gviewer.ps.set_ground_plane_mode("none")
    typical_scale = np.linalg.norm(v, axis=-1).max()
    v /= typical_scale[..., None]
    edge_rad = 0.01
    v_aabb, e_aabb, c_aabb = structures.get_aabb_coords()
    aabb_network = gviewer.ps.register_curve_network(
        "aabb", v_aabb, e_aabb, radius=edge_rad
    )
    aabb_network.add_color_quantity("color", c_aabb, defined_on="edges", enabled=True)
    v_gizmo, e_gizmo, c_gizmo = structures.get_gizmo_coords(0.1)
    gizmo_network = gviewer.ps.register_curve_network(
        "gizmo", v_gizmo, e_gizmo, radius=edge_rad
    )
    gizmo_network.add_color_quantity("color", c_gizmo, defined_on="edges", enabled=True)
    pcs_v = v
    if s_vertices is None:
        gv_scalar = [None]
    else:
        gv_scalar = s_vertices
    if c_vertices is None:
        gv_color = [None]
    else:
        gv_color = c_vertices
    gviewer.ps.set_user_callback(pcs_slider_callback)
    gviewer.ps.set_up_dir("z_up")
    gviewer.register_pointcloud(
        "pc", pcs_v[0], s=gv_scalar[0], c=gv_color[0], radius=0.0006
    )
    gviewer.show()


def meshes_slide_view(
    v, f, s_vertices=None, c_vertices=None, v_vertice=None, s_faces=None, c_faces=None
):
    """
    given some vertices tXVx3 and faces tXFx3
    show the mesh as it changes through time t and allow scrolling through using a slider.
    """
    global meshes_v, meshes_f, gv_scalar, gv_color, gv_vector, gf_scalar, gf_color
    gviewer.init()
    meshes_v = v
    meshes_f = f
    ### next hack is very eww
    if s_vertices is None:
        gv_scalar = [None]
    else:
        gv_scalar = s_vertices
    if c_vertices is None:
        gv_color = [None]
    else:
        gv_color = c_vertices
    if v_vertice is None:
        gv_vector = [None]
    else:
        gv_vector = v_vertice
    if s_faces is None:
        gf_scalar = [None]
    else:
        gf_scalar = s_faces
    if c_faces is None:
        gf_color = [None]
    else:
        gf_color = c_faces

    gviewer.ps.set_user_callback(meshes_slider_callback)
    gviewer.ps.set_up_dir("z_up")
    gviewer.register_mesh(
        "mesh",
        meshes_v[0],
        meshes_f[0],
        s_vertices=gv_scalar[0],
        c_vertices=gv_color[0],
        v_vertices=gv_vector[0],
        s_faces=gf_scalar[0],
        c_faces=gf_color[0],
        edge_width=1.0,
    )
    gviewer.show()


def poses_slide_view(camera_poses):
    """
    given a tensor of t x b x 4 x 4 camera poses, where t is time axis (or step number), b is batch axis, and 4x4 is the c2w transform matrix,
    show the batch of poses and allow scrolling through the time axis using a slider.
    """
    if camera_poses.ndim != 4:
        raise ValueError("camera_poses must be t x b x 4 x 4")
    if camera_poses.shape[2] != 4 or camera_poses.shape[3] != 4:
        raise ValueError("camera_poses must be t x b x 4 x 4")
    global poses
    poses = camera_poses[2:]
    gviewer.init()
    gviewer.ps.set_user_callback(poses_slider_callback)
    gviewer.ps.set_up_dir("z_up")
    gviewer.ps.set_ground_plane_mode("shadow_only")
    edge_rad = 0.0005
    point_rad = 0.002
    v_aabb, e_aabb, c_aabb = structures.get_aabb_coords()
    aabb_network = gviewer.ps.register_curve_network(
        "aabb", v_aabb, e_aabb, radius=edge_rad
    )
    aabb_network.add_color_quantity("color", c_aabb, defined_on="edges", enabled=True)
    v_gizmo, e_gizmo, c_gizmo = structures.get_gizmo_coords(0.1)
    gizmo_network = gviewer.ps.register_curve_network(
        "gizmo", v_gizmo, e_gizmo, radius=edge_rad
    )
    gizmo_network.add_color_quantity("color", c_gizmo, defined_on="edges", enabled=True)
    coa = np.zeros((1, 3))
    gviewer.register_pointcloud(
        "center_of_world",
        coa,
        c=np.array([1.0, 1.0, 1.0])[None, :],
        radius=0.005,
        mode="sphere",
    )
    v_tot, e_tot, c_tot = gviewer.register_camera(
        "poses_init", camera_poses[0], edge_rad, group=True, alpha=0.3
    )
    v_tot, e_tot, c_tot = gviewer.register_camera(
        "poses_orig", camera_poses[1], edge_rad, group=True, alpha=0.3
    )
    gviewer.show()


def poses_static_view(
    camera_poses=None, meshes=None, pointclouds=None, group_cameras=True
):
    """
    visualizes a camera setup
    :param camera_pose: (n, 4, 4) np array of camera to world transforms
    :param meshes: list of (v, f) tuples
    :param pointclouds: list of v
    :param group_cameras: if true, accelerates view but groups camera as a single object
    :return:
    """
    gviewer.init(height=512, width=512)
    gviewer.ps.set_up_dir("z_up")
    edge_rad = 0.0005
    v_aabb, e_aabb, c_aabb = structures.get_aabb_coords()
    aabb_network = gviewer.ps.register_curve_network(
        "aabb", v_aabb, e_aabb, radius=edge_rad
    )
    aabb_network.add_color_quantity("color", c_aabb, defined_on="edges", enabled=True)
    v_gizmo, e_gizmo, c_gizmo = structures.get_gizmo_coords(0.1)
    gizmo_network = gviewer.ps.register_curve_network(
        "gizmo", v_gizmo, e_gizmo, radius=edge_rad
    )
    gizmo_network.add_color_quantity("color", c_gizmo, defined_on="edges", enabled=True)
    coa = np.zeros((1, 3))
    gviewer.register_pointcloud(
        "center_of_world",
        coa,
        c=np.array([1.0, 1.0, 1.0])[None, :],
        radius=0.005,
        mode="sphere",
    )
    if camera_poses is not None:
        v_tot, e_tot, c_tot = gviewer.register_camera(
            "cameras", camera_poses, edge_rad, group_cameras
        )
    if meshes is not None:
        for i, mesh in enumerate(meshes):
            gviewer.register_mesh(
                "mesh_{}".format(i), mesh[0], mesh[1], transparency=0.5
            )
    if pointclouds is not None:
        typical_scale = np.linalg.norm(pointclouds, axis=-1).max()
        for i, pointcloud in enumerate(pointclouds):
            pointcloud /= typical_scale[..., None]
            gviewer.register_pointcloud("pc_{}".format(i), pointcloud, radius=1e-4)
    gviewer.show()


def calibration_static_view(
    camera_pose,
    projector_pose,
    camera_wh,
    projector_wh,
    camera_intrinsics=None,
    camera_distortion=None,
    projector_intrinsics=None,
    forward_map=None,
    fg=None,
    RGB_color_image=None,
    mode="xy",
):
    """
    visualizes a procam pair reconstruction
    :param camera_pose: (4, 4) np array of camera to world transform
    :param projector_pose: (4, 4) np array of projector to world transform
    :param camera_wh: (2,) np array of camera resolution
    :param projector_wh: (2,) np array of projector resolution
    :param camera_intrinsics: (3, 3) np array of camera intrinsics
    :param camera_distortion: (5,) np array of camera distortion
    :param projector_intrinsics: (3, 3) np array of projector intrinsics
    :param forward_map: (h, w, 2) np array of projector to camera mapping
    :param fg: (h, w, 3) np array of projector fg
    :param mode: "xy" or "ij" for forward map last channel encoding
    """
    gviewer.init(height=1024, width=1024)
    gviewer.ps.set_up_dir("z_up")
    edge_rad = 0.0005
    typical_scale = (
        np.linalg.norm(camera_pose[:-1, :]) + np.linalg.norm(projector_pose[:-1, :])
    ) / 2
    camera_pose[:-1, :] /= typical_scale
    projector_pose[:-1, :] /= typical_scale
    v_aabb, e_aabb, c_aabb = structures.get_aabb_coords()
    aabb_network = gviewer.ps.register_curve_network(
        "aabb", v_aabb, e_aabb, radius=edge_rad
    )
    aabb_network.add_color_quantity("color", c_aabb, defined_on="edges", enabled=True)
    v_gizmo, e_gizmo, c_gizmo = structures.get_gizmo_coords(0.1)
    gizmo_network = gviewer.ps.register_curve_network(
        "gizmo", v_gizmo, e_gizmo, radius=edge_rad
    )
    gizmo_network.add_color_quantity("color", c_gizmo, defined_on="edges", enabled=True)
    coa = np.zeros((1, 3))
    gviewer.register_pointcloud(
        "center_of_world",
        coa,
        c=np.array([1.0, 1.0, 1.0])[None, :],
        radius=0.005,
        mode="sphere",
    )
    v_tot, e_tot, c_tot = gviewer.register_camera(
        "camera", camera_pose[None, ...], edge_rad, True, 1.0, 10.0
    )
    v_tot, e_tot, c_tot = gviewer.register_camera(
        "projector", projector_pose[None, ...], edge_rad, True, 1.0, 10.0
    )
    camera_pixels = pixels_in_world_space(camera_wh, camera_intrinsics, camera_pose)
    projector_pixels = pixels_in_world_space(
        projector_wh, projector_intrinsics, projector_pose
    )
    # camera_screen_pc = gviewer.register_pointcloud("camera_screen", camera_pixels, radius=0.0002)
    # projector_screen_pc = gviewer.register_pointcloud("projector_screen", projector_pixels, radius=0.0002)
    (
        points,
        weight_factor,
        cam_origins,
        cam_directions,
        projector_origins,
        projector_directions,
        cam_pixels,
        projector_pixels,
    ) = reconstruct_pointcloud(
        forward_map,
        fg,
        camera_pose,
        projector_pose,
        camera_intrinsics,
        camera_distortion,
        projector_intrinsics,
        mode=mode,
        color_image=RGB_color_image,
        debug=True,
    )
    reconst = gviewer.register_pointcloud(
        "reconstruction", points[:, :3], c=points[:, 3:], radius=0.0002
    )
    t = np.linspace(0, 3, 50)
    cam_ray_points = cam_origins[0] + t[:, None] * cam_directions[0][None, :]
    cam_ray = gviewer.ps.register_curve_network(
        "cam_ray", cam_ray_points, "line", radius=0.0002
    )
    proj_ray_points = (
        projector_origins[0] + t[:, None] * projector_directions[0][None, :]
    )
    proj_ray = gviewer.ps.register_curve_network(
        "proj_ray", proj_ray_points, "line", radius=0.0002
    )
    ray_intersection = points[0:1, :3]
    gviewer.register_pointcloud(
        "ray_intersection",
        ray_intersection,
        c=np.array([1.0, 1.0, 1.0])[None, :],
        radius=0.0005,
        mode="sphere",
    )
    ###
    gviewer.show()



import torch
import numpy as np
from .core import is_np, broadcast_batch


def scalar_projection(a, b):
    """
    computes a scalar projection of vector a onto b
    note: this is the size of the componenet of a tangent to b
    :param a, b: np arrays of size 2 or 3
    """
    return np.dot(a, b) / np.dot(b, b)


def project_point_to_line(p, a, b):
    """
    Returns the closest point on line ab to point p.
    p, a, b: numpy arrays of shape (2,).
    """
    if np.all(a == b):
        return a
    # direction vector of the line
    ab = b - a
    # direction of vector from a to p
    ap = p - a
    # compute the scalar projection of p-a onto b-a
    t = scalar_projection(ap, ab)
    # compute the projection point
    return a + t * ab


def project_point_to_segment(p, a, b):
    """
    Returns the closest point on a finite segment ab to point p.
    p, a, b: numpy arrays of shape (2,).
    """
    if np.all(a == b):
        return a
    # direction vector of the line
    ab = b - a
    # direction of vector from a to p
    ap = p - a
    # compute the scalar projection of p-a onto b-a
    t = scalar_projection(ap, ab)
    # clip result to either a or b
    t = np.clip(t, 0, 1)
    return a + t * ab


def point_line_distance(p, a, b):
    """
    returns the distance between a point p and a line defined by the points a and b
    :param p: point to check (3,) or batch of points to check (B, 3)
    :param a: first vertex of the line (3,)
    :param b: second vertex of the line (3,)
    :return the distance between the point and the line
    """
    nom = np.linalg.norm(np.cross(b - a, p - a), axis=-1)
    denom = np.linalg.norm(b - a, axis=-1)
    return nom / denom


def edge_function(a, b, p):
    """
    returns the "edge function" which equals half the area of the traingle formed by a, b and p
    if the result is positive, p is to the right of the line ab
    :param a: first vertex of the triangle (3,)
    :param b: second vertex of the triangle (3,)
    :param p: point to check (3,) or batch of points to check (B, 3)
    """
    return np.cross(b - a, p - a)


def is_inside_triangle(p, a, b, c):
    """
    checks if a point is inside a triangle
    :param p: point or batch of points to check (3,)
    :param a: first vertex of the triangle (3,)
    :param b: second vertex of the triangle (3,)
    :param c: third vertex of the triangle (3,)
    """
    check1 = np.all(edge_function(a, b, p) >= 0, axis=-1)
    check2 = np.all(edge_function(b, c, p) >= 0, axis=-1)
    check3 = np.all(edge_function(c, a, p) >= 0, axis=-1)
    return check1 & check2 & check3


def triangulate_quad_mesh(F, S=None):
    """
    Triangulates a quad mesh.
    Each quad face (v0, v1, v2, v3) is split into two triangles:
      - Triangle 1: (v0, v1, v2)
      - Triangle 2: (v0, v2, v3)
    :param F: (m,4) np.ndarray of quad face indices.
    :param S: (m, x) some signal per face that needs change according to F
    :return: F_tri (m*2, 3) and S_tri (m*2, x) if S is not None
    """
    m = F.shape[0]
    F_tri = np.empty((m * 2, 3), dtype=F.dtype)
    F_tri[0::2] = F[:, [0, 1, 2]]
    F_tri[1::2] = F[:, [0, 2, 3]]
    if S is None:
        return F_tri
    else:
        S_tri = np.empty((m * 2, S.shape[-1]), dtype=S.dtype)
        S_tri[0::2] = S
        S_tri[1::2] = S
        return F_tri, S_tri


def duplicate_faces(f):
    """
    duplicates *every* face in the mesh, with flipped orientation (and appends it to the end of the tensor)
    note: will transfer to CPU if necessary with current implementation
    :param f: faces of the mesh (Nx3)
    :return: faces of the mesh with duplicated faces
    """
    if is_np(f):
        new_faces = f
    else:
        new_faces = f.detach().cpu().numpy()
    swapped_f = new_faces.copy()
    swapped_f[:, [1, 2]] = swapped_f[:, [2, 1]]
    f_new = np.concatenate([new_faces, swapped_f])
    if not is_np(f):
        f_new = torch.tensor(f_new, dtype=f.dtype, device=f.device)
    return f_new


def remove_duplicate_faces(f):
    """
    remove duplicate faces from a mesh
    note: will transfer to CPU if necessary with current implementation
    :param f: faces of the mesh (Nx3)
    :return: vertices and faces of the mesh without duplicate faces
    """
    if is_np(f):
        f_new = f
    else:
        f_new = f.detach().cpu().numpy()
    f_new = np.sort(f_new, axis=1)
    f_new = np.unique(f_new, axis=0)
    if not is_np(f):
        f_new = torch.tensor(f_new, dtype=f.dtype, device=f.device)
    return f_new


def get_aspect_ratio(v: torch.Tensor, f: torch.Tensor):
    """
    measure aspect ratio of all triangles using: (circumradius / 2inradius)
    aspect_ratio < 1 for illegal or degenerate triangle
    aspect_ratio = 1 for equilateral triangle
    aspect_ratio > 1 everything else
    v: vertices (V,3)
    f: faces (F,3)
    """
    v0, v1, v2 = v[f].unbind(dim=-2)
    a = (v0 - v1).norm(dim=-1)
    b = (v0 - v2).norm(dim=-1)
    c = (v1 - v2).norm(dim=-1)
    s = (a + b + c) / 2
    return a * b * c / (8 * (s - a) * (s - b) * (s - c))


def get_face_areas(v, f, eps=1e-7):
    """
    :param v: vertex tensor Vx3
    :param f: face tensor Fx3
    :param eps: area thresold for degenerate face
    :return: tensor of face normals
    """
    face_vertices = v[f]
    v0 = face_vertices[:, 0, :]
    v1 = face_vertices[:, 1, :]
    v2 = face_vertices[:, 2, :]
    cross = torch.cross((v1 - v0), (v2 - v0), dim=1)
    norms = torch.norm(cross, dim=1, keepdim=True)
    face_areas = norms / 2
    face_areas[face_areas < eps] = eps
    return face_areas


def normalize_vertices(vertices, mode="unit_sphere"):
    """
    shift and resize mesh to fit into a bounding volume
    """
    eps = 1e-7
    if is_np(vertices):
        vertices -= (vertices.min(axis=0) + vertices.max(axis=0)) / 2
        if mode == "unit_sphere":
            vertices = vertices / (np.linalg.norm(vertices, axis=-1).max() + eps)
        elif mode == "unit_cube":
            vertices = vertices / (np.abs(vertices).max(dim=-1) + eps)
            raise NotImplementedError
    else:
        if mode == "unit_sphere":
            vertices -= (vertices.min(dim=0)[0] + vertices.max(dim=0)[0]) / 2
            vertices = vertices / (torch.norm(vertices, dim=-1).max() + eps)
        elif mode == "unit_cube":
            raise NotImplementedError
    return vertices


def calc_face_normals(
    vertices: torch.Tensor, faces: torch.Tensor, normalize: bool = False
):
    """
    # V,3 first vertex may be unreferenced
    # F,3 long, first face may be all zeros

         n
         |
         c0     corners ordered counterclockwise when
        / \     looking onto surface (in neg normal direction)
      c1---c2
    """
    full_vertices = vertices[faces]  # F,(3,3)
    v0, v1, v2 = full_vertices.unbind(dim=1)  # F,3
    face_normals = torch.cross(v1 - v0, v2 - v0, dim=1)  # F,3
    if normalize:
        face_normals = torch.nn.functional.normalize(face_normals, eps=1e-6, dim=1)
    return face_normals  # F,3


def calc_vertex_normals(
    vertices: torch.Tensor, faces: torch.Tensor, face_normals: torch.Tensor = None
):
    """
    # V,3 first vertex may be unreferenced
    # F,3 long, first face may be all zero
    return # F,3, not normalized
    """
    F = faces.shape[0]
    if face_normals is None:
        face_normals = calc_face_normals(vertices, faces)
    vertex_normals = torch.zeros(
        (vertices.shape[0], 3, 3), dtype=vertices.dtype, device=vertices.device
    )  # V,C=3,3
    vertex_normals.scatter_add_(
        dim=0,
        index=faces[:, :, None].expand(F, 3, 3),
        src=face_normals[:, None, :].expand(F, 3, 3),
    )
    vertex_normals = vertex_normals.sum(dim=1)  # V,3
    return torch.nn.functional.normalize(vertex_normals, eps=1e-6, dim=1)


def get_face_centroids(v, f):
    return torch.mean(v[f], dim=-2)


def faces2edges_naive(faces: np.array):
    """
    a naive implementation in numpy to extract edges from a triangular / quad mesh.
    note: assumes manifold mesh with no borders
    :param faces: (n, 3) or (n, 4) np.uint32 of face indices
    :return edges (m, 2), f2e (n, 3 or 4) and e2f(m, 2)
    """
    edge_list = []
    faces_to_edges = []

    for face in faces:
        face_edges = []
        num_vertices = len(face)
        for i in range(num_vertices):
            v0, v1 = sorted((face[i], face[(i + 1) % num_vertices]))
            edge = (v0, v1)
            edge_list.append(edge)
            face_edges.append(edge)
        faces_to_edges.append(face_edges)

    # Convert to numpy array and find unique edges
    edge_array = np.array(edge_list, dtype=np.int32)
    unique_edges, edge_indices = np.unique(edge_array, axis=0, return_inverse=True)

    # Map faces to their corresponding unique edges
    faces_to_edges = np.array(edge_indices).reshape(faces.shape)

    # initialize the edges_to_faces array with a default value and fill it in:
    num_unique_edges = unique_edges.shape[0]
    edges_to_faces = -np.ones((num_unique_edges, 2), dtype=np.int32)
    for face_idx, face_edge_indices in enumerate(faces_to_edges):
        for edge_idx in face_edge_indices:
            if edges_to_faces[edge_idx, 0] == -1:
                edges_to_faces[edge_idx, 0] = face_idx
            else:
                edges_to_faces[edge_idx, 1] = face_idx
    return unique_edges, faces_to_edges, edges_to_faces


def calc_edges(faces: torch.Tensor, with_edge_to_face: bool = False, with_dummies=True):
    """
    # F,3 long - first face may be dummy with all zeros
    returns tuple of
    - edges E,2 long, 0 for unused, lower vertex index first
    - face_to_edge F,3 long
    - (optional) edge_to_face shape=E,[left,right],[face,side]

    o-<-----e1     e0,e1...edge, e0<e1
    |      /A      L,R....left and right face
    |  L /  |      both triangles ordered counter clockwise
    |  / R  |      normals pointing out of screen
    V/      |
    e0---->-o
    """

    F = faces.shape[0]
    # make full edges, lower vertex index first
    face_edges = torch.stack((faces, faces.roll(-1, 1)), dim=-1)  # F*3,3,2
    full_edges = face_edges.reshape(F * 3, 2)
    sorted_edges, _ = full_edges.sort(dim=-1)  # F*3,2 todo min/max faster?

    # make unique edges
    edges, full_to_unique = torch.unique(
        input=sorted_edges, return_inverse=True, dim=0
    )  # (E,2),(F*3)
    E = edges.shape[0]
    face_to_edge = full_to_unique.reshape(F, 3)  # F,3
    if not with_edge_to_face:
        return edges, face_to_edge

    is_right = full_edges[:, 0] != sorted_edges[:, 0]  # F*3
    edge_to_face = torch.zeros(
        (E, 2, 2), dtype=torch.long, device=faces.device
    )  # E,LR=2,S=2
    scatter_src = torch.cartesian_prod(
        torch.arange(0, F, device=faces.device), torch.arange(0, 3, device=faces.device)
    )  # F*3,2
    edge_to_face.reshape(2 * E, 2).scatter_(
        dim=0,
        index=(2 * full_to_unique + is_right)[:, None].expand(F * 3, 2),
        src=scatter_src,
    )  # E,LR=2,S=2
    if with_dummies:
        edge_to_face[0] = 0
    return edges, face_to_edge, edge_to_face


def calc_edge_length(vertices: torch.Tensor, edges: torch.Tensor):
    """
    # V,3 first may be dummy
    return # E,2 long, lower vertex index first, (0,0) for unused
    """
    full_vertices = vertices[edges]  # E,2,3
    a, b = full_vertices.unbind(dim=1)  # E,3
    return torch.norm(a - b, p=2, dim=-1)


def prepend_dummies(vertices: torch.Tensor, faces: torch.Tensor):
    """
    prepend dummy elements to vertices and faces to enable "masked" scatter operations
    :param vertices (V,3)
    :param faces (F,3) long
    """
    V, D = vertices.shape
    vertices = torch.concat(
        (torch.full((1, D), fill_value=torch.nan, device=vertices.device), vertices),
        dim=0,
    )
    faces = torch.concat(
        (torch.zeros((1, 3), dtype=torch.long, device=faces.device), faces + 1), dim=0
    )
    return vertices, faces


def remove_dummies(vertices: torch.Tensor, faces: torch.Tensor):
    """
    remove dummy elements added with prepend_dummies()
    :param vertices (V,3) - first vertex all nan and unreferenced
    :param faces (F,3) long - first face all zeros
    :return: V-1,D and F-1,3
    """
    return vertices[1:], faces[1:] - 1


def ray_sphere_intersection(sphere_origin, sphere_radius, ray_origin, ray_direction):
    """
    returns the distance along a ray to the intersection point between a sphere and a ray
    # sphere_origin, sphere_radius, ray_origin, ray_direction: # B,3
    # returns: # B
    """
    radius = sphere_radius
    radius2 = radius**2
    center = sphere_origin
    L = center - ray_origin
    tca = torch.bmm(L[:, None, :], ray_direction[:, :, None]).squeeze()
    tca[tca < 0] = float("Inf")
    d2 = torch.bmm(L[:, None, :], L[:, :, None]).squeeze() - tca * tca
    d2[d2 > radius2] = -float("inf")
    thc = torch.sqrt(radius2 - d2)
    t0 = tca - thc
    t1 = tca + thc
    mask = t0 > t1
    t0[mask] = t1[mask]
    return t0


def qslerp(qa, qb, t):
    """
    interpolates between two quanternions
    :param qa: first quaternions
    :param qb: second quaternions
    :param t: number between 0 and 1 indicating the interpolation factor
    :return: b x 4 array of quaternions
    """
    qm = np.zeros_like(qa)
    cosHalfTheta = qa.dot(qb)
    if abs(cosHalfTheta) >= 1.0:  # theta = 0 degrees and we can return qa
        qm = qa
        return qm
    halfTheta = np.arccos(cosHalfTheta)
    sinHalfTheta = np.sqrt(1.0 - cosHalfTheta * cosHalfTheta)
    if (
        np.fabs(sinHalfTheta) < 0.001
    ):  # if theta = 180 degrees then result is not fully defined
        qm = qa * 0.5 + qb * 0.5
        return qm
    ratioA = np.sin((1 - t) * halfTheta) / sinHalfTheta
    ratioB = np.sin(t * halfTheta) / sinHalfTheta
    qm = qa * ratioA + qb * ratioB
    return qm


def ray_ray_intersection(oa, da, ob, db):
    """
    returns points closest to each corresponding ray of form o+t*d in the bundle
    :param oa: origin of ray a (b x 3)
    :param da: direction of ray a (b x 3)
    :param ob: origin of ray b (b x 3)
    :param db: direction of ray b (b x 3)
    :return: points closest to each ray (b x 3) and a weight factor that goes to 0 if the lines are parallel
    """
    batched_input = True
    if oa.ndim == 1:
        oa = oa[None, :]
        batched_input = False
    if da.ndim == 1:
        da = da[None, :]
        batched_input = False
    if ob.ndim == 1:
        ob = ob[None, :]
        batched_input = False
    if db.ndim == 1:
        db = db[None, :]
        batched_input = False
    oa, ob, da, db = broadcast_batch(oa, ob, da, db)
    da = da / np.linalg.norm(da, axis=1, keepdims=True)
    db = db / np.linalg.norm(db, axis=1, keepdims=True)
    c = np.cross(da, db)
    denom = np.linalg.norm(c, axis=-1) ** 2
    t = ob - oa
    stackedb = np.concatenate([t[:, None, :], db[:, None, :], c[:, None, :]], axis=1)
    stackeda = np.concatenate([t[:, None, :], da[:, None, :], c[:, None, :]], axis=1)
    ta = np.linalg.det(stackedb) / (denom + 1e-10)
    tb = np.linalg.det(stackeda) / (denom + 1e-10)
    ta[ta < 0] = 0
    tb[tb < 0] = 0
    points = (oa + ta[:, None] * da + ob + tb[:, None] * db) * 0.5
    if not batched_input:
        points = points[0]
    return points, denom


def get_center_of_attention(c2w):
    """
    find a central point of a batch of c2w transforms 4x4 they are all looking at.
    note: point is weighted by distance between lines.
    :param camera_poses: n x 4 x 4 np array of c2w matrices
    :return: the center of attention in 3d world coordinates
    """
    if c2w.ndim != 3:
        raise ValueError("c2w must be a 3d array of 4x4 matrices")
    if c2w.shape[1] != 4 or c2w.shape[2] != 4:
        raise ValueError("c2w must be a 3d array of 4x4 matrices")
    totw = 0.0
    totp = np.array([0.0, 0.0, 0.0])
    ps = []
    rays = []
    for f in c2w:
        mf = f[0:3, :]
        for g in c2w:
            mg = g[0:3, :]
            p, w = ray_ray_intersection(mf[:, 3], mf[:, 2], mg[:, 3], mg[:, 2])
            if w > 0.01:
                totp += p * w
                totw += w
                # ps.append(p)
                # rays.append([mf[:, 3], mf[:, 2], mg[:, 3], mg[:, 2]])
    totp /= totw
    return totp  # np.array(rays), np.array(ps)


def scale_poses(c2w, n=1.0):
    """
    scales c2w transforms so that the average distance to the center of attention is n
    :param c2w: n x 4 x 4 np array of c2w matrices
    :param n: the average distance to the center of attention
    :return: the scaled c2w matrices
    """
    avglen = np.mean(np.linalg.norm(c2w[:, 0:3, 3], axis=-1))
    # print("avg camera distance from origin", avglen)
    c2w[:, 0:3, 3] *= n / avglen
    return c2w, avglen


def merge_meshes(v1, f1, v2=None, f2=None):
    """merge two meshes into one"""
    if v2 is None or f2 is None:
        return v1, f1
    v = np.concatenate([v1, v2], axis=0)
    f = np.concatenate([f1, f2 + v1.shape[0]], axis=0)
    return v, f


def find_princple_componenets(v: torch.Tensor):
    """
    finds the principle components of a points nx3
    :param v: nx3 torch tensor of points
    :return: the principle components (column major)
    """
    if type(v) == torch.Tensor:
        cov = v.T @ v
        _, vecs = torch.linalg.eig(cov)
        if torch.imag(vecs).any():
            raise ValueError("imaginary eigenvectors")
        vecs = torch.real(vecs)
        if torch.det(vecs) < 0:
            swap_axis = torch.tensor(
                [[1, 0, 0], [0, 0, 1], [0, 1, 0]], dtype=v.dtype, device=v.device
            )
            vecs = swap_axis @ vecs
    else:
        raise TypeError("v must be a torch tenor")
    return vecs


def remove_unreferenced_vertices(v, f):
    """
    removes unreferenced vertices from a mesh
    :param v: Vx3 np array of vertices
    :param f: Fx3 np array of faces
    :return: the new vertices and faces, and a map from new to old indices (-1 if unreferenced)
    """
    referenced = np.zeros((v.shape[0]), dtype=bool)
    referenced[f.flatten()] = True
    idx = -1 * np.ones((v.shape[0]), dtype=np.int64)
    idx[referenced] = np.arange(np.sum(referenced), dtype=np.int64)
    f = idx[f.flatten()].reshape((-1, f.shape[1]))
    v = v[referenced]
    return v, f, idx


def get_edges(f: np.ndarray):
    """
    given a numpy array of faces of a triangular mesh F x 3, returns a numpy array of edges
    :param f: F x 3 numpy array of faces
    :return: E x 2 numpy array of edges
    """
    e1 = np.concatenate((f[:, 0:1], f[:, 1:2]), axis=1)
    e2 = np.concatenate((f[:, 1:2], f[:, 2:3]), axis=1)
    e3 = np.concatenate((f[:, 2:3], f[:, 0:1]), axis=1)
    e = np.concatenate((e1, e2, e3), axis=0)
    e = np.sort(e, axis=-1)
    e = np.unique(e, axis=0)
    return e


def edge_contraction(v, f, edge_to_contract, new_v_location):
    """
    contracts an edge in a mesh and moves the vertex to a new location
    :param v: V x 3 numpy array of vertices
    :param f: F x 3 numpy array of faces
    :param edge_to_contract: 2, numpy array of the edge to contract
    :param new_v_location: 3, numpy array of the new vertex location
    :return: the contracted mesh
    """
    v1 = edge_to_contract[0]
    v2 = edge_to_contract[1]
    # replace all instances of v1 with v2 in f
    f[f == v1] = v2
    # remove degenerate triangles
    degenerate_mask = (np.diff(np.sort(f, axis=-1)) == 0).any(axis=-1)
    f = f[~degenerate_mask]
    # place v2 in v_hat
    v[v2] = new_v_location
    # remove v1 from v
    v, f, _ = remove_unreferenced_vertices(v, f)
    return v, f


def clean_infinite_vertices(v, f):
    """
    removes vertices that are infinite / nan, and all their incident faces
    :param v: V x 3 numpy array of vertices
    :param f: F x 3 numpy array of faces
    :return: the cleaned mesh
    """
    finite_mask = np.isfinite(v).any(axis=-1)  # finite mask
    if ~np.all(finite_mask):
        f = f[finite_mask[f].all(axis=-1)]  # remove faces with infinite vertices
        v, f, _ = remove_unreferenced_vertices(v, f)  # update vertex indices
    return v, f



import torch
import numpy as np
from PIL import Image


def is_np(x):
    """
    checks if x is a numpy array or torch tensor (will raise an error if x is neither)
    :param x: object to check
    :return: True if x is a numpy array, False if x is a torch tensor
    """
    if type(x) == np.ndarray:
        return True
    elif type(x) == torch.Tensor:
        return False
    else:
        raise ValueError("input must be torch.Tensor or np.ndarray")


def to_hom(x):
    """
    converts a vector to homogeneous coordinates (concatenates 1 along last dimension)
    :param x: (..., c) numpy array or torch tensor
    :return: (..., c+1) numpy array or torch tensor
    """
    if is_np(x):
        if x.ndim == 1:
            return np.concatenate((x, np.array([1], dtype=x.dtype)))
        else:
            return np.concatenate(
                (x, np.ones((*x.shape[:-1], 1), dtype=x.dtype)), axis=-1
            )
    else:
        if x.ndim == 1:
            return torch.cat((x, torch.ones(1, device=x.device)))
        else:
            return torch.cat((x, torch.ones(*x.shape[:-1], 1, device=x.device)), dim=-1)


def homogenize(x, keepdim=False):
    """
    normalizes a homogeneous vector by dividing by the last coordinate
    :param x: nx3 numpy array
    :return: nx4 numpy array
    """
    x = x / x[..., -1:]
    if not keepdim:
        x = x[..., :-1]
    return x


def normalize(x, eps=1e-7):
    """
    normalizes a vector by dividing by its norm
    :param x: (... , c) numpy array
    :return: (... , c) numpy array normalized along the last dimension
    """
    if is_np(x):
        return x / (np.linalg.norm(x, axis=-1, keepdims=True) + eps)
    else:
        return x / (torch.norm(x, dim=-1, keepdim=True) + eps)


def broadcast_batch(*args):
    """
    broadcast a list of arrays to the same shape on the batch dimension
    assumes first dimension is batch unless ndim = 1 for all inputs (but then does not broadcast)
    :param args: list of arrays
    :return: list of arrays with the same shape
    """
    shapes = [a.shape for a in args]
    ndims = np.array([a.ndim for a in args])
    new_args = []
    if (ndims <= 1).all():
        for a in args:
            new_args.append(a[None, :])
        return new_args
    elif ndims.min() == 1:
        raise ValueError("cannot broadcast 1d and nd arrays")
    else:
        batch_dim = np.array([s[0] for s in shapes]).max()
        for i, a in enumerate(args):
            if a.shape[0] == batch_dim:
                new_args.append(a)
            else:
                new_args.append(np.broadcast_to(a, (batch_dim, *a.shape[1:])))
    return new_args


def repeat(x, n):
    """
    repeats a numpy array or torch tensor
    :param x: the input
    :param n: a tuple describing the repeat per dimension of x
    return the repeated input
    """
    if is_np(x):
        y = np.tile(x, n)
    else:
        y = torch, repeat(x, *n)
    return y


def compose_rt(R: np.array, t: np.array, square=False):
    """
    composes a n x 3 x 4 numpy array from rotation and translation.
    will broadcast upon batch dimension if necessary.
    :param R: nx3x3 numpy array
    :param t: nx3 numpy array
    :param square: if True, output will be 4x4, otherwise 3x4
    :return: n x 3 x 4 composition of the rotation and translation
    """
    RR, tt = broadcast_batch(R, t)
    Rt = np.concatenate((RR, tt[:, :, None]), axis=-1)
    if square:
        Rt = to_44(Rt)
    return Rt


def to_44(mat):
    """
    converts a (3,4) to a (4,4) matrix by concatenating [0, 0, 0, 1]
    :param mat: (n, 3, 4) numpy array (n can be any number of including 0)
    :return: (n, 4, 4) numpy array
    """
    if mat.shape[-2:] == (4, 4):
        return mat
    if mat.shape[-2:] != (3, 4):
        raise ValueError("mat must be 3x4")
    if is_np(mat):
        to_cat = np.broadcast_to(
            np.array([0, 0, 0, 1]).astype(mat.dtype), (*mat.shape[:-2], 1, 4)
        )
        new_mat = np.concatenate((mat, to_cat), axis=-2)
    else:
        to_cat = torch.zeros(
            (*mat.shape[:-2], 1, 4), dtype=mat.dtype, device=mat.device
        )
        to_cat[..., -1] = 1
        new_mat = torch.cat((mat, to_cat), dim=-2)
    return new_mat


def to_34(mat: np.array):
    """
    converts a (n, 4, 4) to a (n, 3, 4) matrix by removeing the last row
    :param mat: (n, 4, 4) numpy array
    :return: (n, 3, 4) numpy array
    """
    if mat.ndim == 3:
        if mat.shape[1:] != (4, 4):
            raise ValueError("mat must be 4x4")
        return mat[:, :-1, :]
    else:
        if mat.shape != (4, 4):
            raise ValueError("mat must be 4x4")
        return mat[:-1, :]


def to_np(arr: torch.Tensor):
    """
    converts a tensor to numpy array
    :param arr: tensor
    :return: numpy array
    """
    if type(arr) == torch.Tensor:
        return arr.detach().cpu().numpy()
    elif type(arr) == np.ndarray:
        return arr
    elif type(arr) == Image.Image:
        return np.array(arr)
    elif type(arr) == list:
        return np.array(arr)
    else:
        raise TypeError("cannot convert {} to numpy array".format(str(type(arr))))


def to_numpy(arr: torch.Tensor):
    """
    converts a tensor to numpy array
    :param arr: tensor
    :return: numpy array
    """
    return to_np(arr)


def to_torch(arr: np.array, device="cpu", dtype=None):
    """
    converts a numpy array to a torch tensor
    :param arr: numpy array
    :param dtype: dtype of the tensor
    :param device: device to put the tensor on
    :return: torch tensor
    """
    if dtype is None:
        return torch.tensor(arr, device=device)
    else:
        return torch.tensor(arr, dtype=dtype, device=device)


def to_8b(x, clip=True):
    """
    convert an array (float, double) array to 8 bit
    :param x: array
    :param clip: if True, clips values to [0,1]
    :return: 8 bit array
    """
    if is_np(x):
        if x.dtype == np.float16 or x.dtype == np.float32 or x.dtype == np.float64:
            if clip:
                x = np.clip(x, 0, 1)
            return (255 * x).round().astype(np.uint8)
        elif x.dtype == bool:
            return x.astype(np.uint8) * 255
        elif x.dtype == np.uint8:
            return x
    else:
        if x.dtype == torch.float32 or x.dtype == torch.float64:
            if clip:
                x = torch.clamp(x, 0, 1)
            return (255 * x).round().type(torch.uint8)
        elif x.dtype == torch.bool:
            return x.type(torch.uint8) * 255
        elif x.dtype == torch.uint8:
            return x


def to_float(x, clip=True):
    """
    convert a 8bit or bool array to float
    :param x: array
    :param clip: if True, clips values to [0,1]
    :return: float array
    """
    if is_np(x):
        if x.dtype == np.uint8:
            return x.astype(np.float32) / 255
        elif x.dtype == np.float32:
            if clip:
                x = np.clip(x, 0, 1)
            return x
        elif x.dtype == bool:
            return x.astype(np.float32)
        else:
            raise ValueError("unsupported dtype")
    else:
        if x.dtype == torch.uint8:
            return x.to(torch.float32) / 255
        elif x.dtype == torch.float32:
            if clip:
                x = torch.clamp(x, 0, 1)
            return x
        elif x.dtype == torch.bool:
            return x.to(torch.float32)
        else:
            raise ValueError("unsupported dtype")


def to_PIL(x: np.array):
    """
    convert a numpy float array to a PIL image
    :param x: array
    :return: PIL image
    """
    if x.ndim == 3:
        return Image.fromarray(to_8b(x))
    elif x.ndim == 2:
        return Image.fromarray(to_8b(x[:, :, None]), mode="L")
    else:
        raise ValueError("unsupported array dimensions")


def map_range(x, in_min, in_max, out_min, out_max):
    """
    given an array, an input and output range, maps the input from the input range to the output range
    :param x: input
    :param in_min: input range minimum
    :param in_max: input range maximum
    :param out_min: output range minimum
    :param out_max: output range maximum
    :return: mapped input
    """
    x_01 = (x - in_min) / (in_max - in_min)
    return x_01 * (out_max - out_min) + out_min


def map_to_01(x):
    """
    maps an input to [0,1]
    :param x: input
    :return: mapped input
    """
    return map_range(x, x.min(), x.max(), 0, 1)


def swap_columns(x, col1_index, col2_index):
    """
    swaps two columns of a numpy array inplace
    :param x: array N x C
    :param col1_index: index of the first column
    :param col2_index: index of the second column
    :return: array with swapped columns
    """
    x[..., [col2_index, col1_index]] = x[..., [col1_index, col2_index]]
    return x


def color_to_gray(x, keep_channels=False):
    """
    convert image to gray scale by averaging over channels
    :param x: numpy array or torch tensor (n, h, w, c) or (h, w, c) where c >= 1
    :keep_channels: if True, will perserve c by repeating the array after average
    :return: the gray scale version of x
    """
    orig_ndim = x.ndim
    if (orig_ndim != 3) and (orig_ndim != 4):
        raise ValueError("ndim of x must be 3 (h, w, c) or 4 (n, h, w, c)")
    c = x.shape[-1]
    orig_dtype = x.dtype
    if is_np(x):
        if x.dtype == np.float32 or x.dtype == np.float64:
            x = x.mean(axis=-1, keepdims=True)
        else:
            x = x.astype(np.float32).mean(axis=-1, keepdims=True)
            x = x.round().astype(orig_dtype)
        if keep_channels:
            if orig_ndim == 3:
                x = np.tile(x, (1, 1, c))
            else:
                x = np.tile(x, (1, 1, 1, c))
    else:
        if x.dtype == torch.float32 or x.dtype == torch.float64:
            x = x.mean(dim=-1, keepdim=True)
        else:
            x = x.to(torch.float32).mean(dim=-1, keepdim=True)
            x = x.round().to(orig_dtype)
        if keep_channels:
            if orig_ndim == 3:
                x = x.repeat(1, 1, c)
            else:
                x = x.repeat(1, 1, 1, c)
    return x



import numpy as np
from . import structures
from .core import to_hom, homogenize
from .gsoup_io import write_to_json


class PolyscopeStub:
    def __getattr__(self, name):
        def wrapper(*args, **kwargs):
            print("{} was called".format(name))

        return wrapper


class gviewer:
    """
    wrapper class for polyscope, invoked as a singelton ("from gsoup.viewer import gviewer")
    offers a stub for when polyscope is not installed, window resizing and a few other things
    """

    def __init__(self):
        self.is_init = False

    def init(
        self,
        up_dir="z_up",
        look_at=None,
        ground_mode="shadow_only",
        projection_mode="perspective",
        width=None,
        height=None,
    ):
        """
        given a ps instance, intializes it
        :param up_dir: up direction
        :param look_at: where to look at (3 tuples eye, at ,up)
        :param ground_mode: ground render mode
        :param projection_mode: projection_mode
        """
        if width is not None:
            if height is None:
                height = width
            poly_dict = {
                "windowHeight": height,
                "windowPosX": 50,
                "windowPosY": 50,
                "windowWidth": width,
            }
            write_to_json(poly_dict, "./.polyscope.ini")
        try:
            import polyscope as ps
            import polyscope.imgui as psim

            ps.init()
        except RuntimeError:
            ps = PolyscopeStub()
            psim = PolyscopeStub()
        self.ps = ps
        self.clean()
        self.psim = psim
        self.ps.set_up_dir(up_dir)
        if look_at is None:
            eye, at, up = (-1.0, 0.75, -1.0), (0.0, 0.0, 0.0), (0.0, 1.0, 0.0)
        else:
            eye, at, up = look_at
        self.ps.look_at_dir(eye, at, up)
        self.ps.set_ground_plane_mode(ground_mode)
        self.ps.set_view_projection_mode(
            projection_mode
        )  # ps.set_view_projection_mode("orthographic")
        self.is_init = True

    def register_pointcloud(
        self,
        name,
        points,
        c=None,
        s=None,
        v=None,
        radius=1e-2,
        mode="quad",
        enabled=True,
    ):
        """
        register a point cloud to polyscope
        """
        ps_pointcloud = self.ps.register_point_cloud(
            name,
            points,
            radius=radius,
            point_render_mode=mode,
            enabled=enabled,
        )  # color=np.array([0.5, 0.1, 0.3])  # sphere / quad
        if c is not None:
            ps_pointcloud.add_color_quantity("colors", c, enabled=True)
        if s is not None:
            s = (s - np.min(s)) / (np.max(s) - np.min(s))
            s = np.nan_to_num(s, nan=1.0, posinf=1.0, neginf=1.0)
            ps_pointcloud.add_scalar_quantity(
                "scalar_value", s, enabled=True, vminmax=(0.0, 1.0), cmap="reds"
            )
        if v is not None:
            # v /= (np.linalg.norm(v, axis=-1) + 1e-6)
            ps_pointcloud.add_vector_quantity(
                "vecs", v, enabled=True, radius=0.002, length=0.1, color=(0.2, 0.5, 0.5)
            )
        return ps_pointcloud

    def register_camera(self, name, poses, edge_rad, group=True, alpha=1.0, scale=0.1):
        """
        register a camera structure to polyscope
        """
        v_cam, e_cam, c_cam = structures.get_camera_coords(scale)
        v_tot = []
        e_tot = []
        c_tot = []
        for i, pose in enumerate(poses):
            v = homogenize((pose @ to_hom(v_cam).T).T)
            if group:
                v_tot.append(v)
                c_tot.append(c_cam)
                e_tot.append(e_cam + (i * len(v_cam)))
            else:
                ps_net = self.ps.register_curve_network(
                    "{}_{}".format(name, i), v, e_cam, radius=edge_rad
                )
                ps_net.add_color_quantity(
                    "color", c_cam, defined_on="edges", enabled=True
                )
                ps_net.set_transparency(alpha)
        if group:
            v_tot = np.array(v_tot).reshape(-1, 3)
            e_tot = np.array(e_tot).reshape(-1, 2)
            c_tot = np.array(c_tot).reshape(-1, 3)
            ps_net = self.ps.register_curve_network(name, v_tot, e_tot, radius=edge_rad)
            ps_net.add_color_quantity("color", c_tot, defined_on="edges", enabled=True)
        ps_net.set_transparency(alpha)
        return v_tot, e_tot, c_tot

    def register_mesh(
        self,
        name,
        v,
        f,
        transparency=1.0,
        edge_width=0.0,
        color=[0.5, 0.5, 0.5],
        smooth_shade=True,
        s_vertices=None,
        c_vertices=None,
        v_vertices=None,
        s_faces=None,
        c_faces=None,
        cmap=None,
        enabled=True,
    ):
        """
        regiter a mesh to polyscope
        :param ps: polyscope instance
        :param name: name of the mesh
        :param v: vertices
        :param f: faces
        :param transparency: transparency of the mesh
        :param edge_width: edge width of the mesh
        :param smooth_shade: smooth shading
        :param s_vertices: vertex scalar values
        :param c_vertices: vertex scalar values
        :param v_vertices: vertex vectors
        :param s_faces: face scalar values
        :param c_faces: face colors
        """
        ps_mesh = self.ps.register_surface_mesh(
            name,
            v,
            f,
            edge_width=edge_width,
            transparency=transparency,
            color=color,
            smooth_shade=smooth_shade,
            enabled=enabled,
        )
        if cmap is None:
            cmap = "reds"
        if s_vertices is not None:
            s = (s_vertices - np.min(s_vertices)) / (
                np.max(s_vertices) - np.min(s_vertices)
            )
            ps_mesh.add_scalar_quantity(
                "scalar",
                s,
                defined_on="vertices",
                enabled=True,
                vminmax=(0.0, 1.0),
                cmap=cmap,
            )
        if c_vertices is not None:
            ps_mesh.add_color_quantity(
                "color", c_vertices, defined_on="vertices", enabled=True
            )
        if v_vertices is not None:
            ps_mesh.add_vector_quantity(
                "vecs",
                v_vertices,
                enabled=True,
                # vectortype='ambient',
                radius=0.01,
                length=0.1,
                color=(0.2, 0.5, 0.5),
            )
        if s_faces is not None:
            s = (s_faces - np.min(s_faces)) / (np.max(s_faces) - np.min(s_faces))
            ps_mesh.add_scalar_quantity(
                "fscalar",
                s,
                defined_on="faces",
                enabled=True,
                vminmax=(0.0, 1.0),
                cmap=cmap,
            )
        if c_faces is not None:
            ps_mesh.add_color_quantity(
                "fcolor", c_faces, defined_on="faces", enabled=True
            )
        return ps_mesh

    def show(self):
        self.ps.show()

    def clean(self):
        self.ps.remove_all_structures()

    def clear(self):
        self.ps.clean()

    def remove_all_structures(self):
        self.ps.clean()


gviewer = gviewer()



import numpy as np
import torch
from PIL import Image, ImageDraw, ImageFont
from scipy import interpolate, spatial
from scipy.stats import wasserstein_distance
from .core import (
    to_8b,
    to_float,
    to_hom,
    homogenize,
    broadcast_batch,
    is_np,
    to_torch,
    to_np,
)
from .structures import get_gizmo_coords
from pathlib import Path


def add_alpha(images, alphas):
    """
    adds an alpha channel to a batch of images
    :param images: numpy image b x h x w x 3 or h x w x 3
    :param alpha: alpha channel b x h x w x 1 or h x w x 1
    :return: numpy image b x h x w x (c+1) or h x w x (c+1)
    """
    if images.shape[-1] != 3:
        raise ValueError("images must have 3 channels")
    if images.ndim != 3 and images.ndim != 4:
        raise ValueError("image must be a 3D/4D array")
    if alphas.ndim != 3 and alphas.ndim != 4:
        raise ValueError("alphas must be a 3D/4D array")
    if images.ndim == 3 and alphas.ndim == 4:
        raise ValueError("images and alphas must have the same number of dimensions")
    if images.ndim == 4:
        if images.shape[0] != alphas.shape[0] and alphas.shape[0] != 1:
            raise ValueError(
                "images and alpha must have the same batch size, or alphas batch size must equal 1"
            )
        if images.shape[1:3] != alphas.shape[1:3]:
            raise ValueError("images and alphas must have the same spatial size")
        images, alphas = broadcast_batch(images, alphas)
    else:
        # both are ndim=3
        if images.shape[:2] != alphas.shape[:2]:
            raise ValueError("images and alphas must have the same spatial size")
    if is_np(images):
        return np.concatenate((images, alphas), axis=-1)
    else:
        return torch.cat((images, alphas), dim=-1)


def alpha_compose(images, backgrounds=None, bg_color=None):
    """
    composes a single or batch of RGBA images into a single or batch of RGB images.
    if backgrounds is provided, will blend them with the images, otherwise will blend with bg_color.
    if no backgrounds or bg_color is provided, the background is assumed to be black.
    :param image: b x H x W x 4 or H x W x 4
    :param background: b x H x W x 3 or H x W x 3
    :param bg_color: 3 or b x 3 float32 array
    :return: b x H x W x 3 or H x W x 3
    """
    if images.ndim != 3 and images.ndim != 4:
        raise ValueError("image must be 3 or 4 dimensional")
    if images.shape[-1] != 4:
        raise ValueError("image must have 4 channels")
    if backgrounds is not None:
        if images.shape[:-1] != backgrounds.shape[:-1]:
            raise ValueError("backgrounds must have same shape as images")
    if is_np(images):
        if bg_color is None:
            bg_color = np.array([0.0, 0.0, 0.0]).astype(np.float32)
        if images.dtype != np.float32:
            images = to_float(images)
    else:
        if bg_color is None:
            bg_color = torch.tensor(
                [0.0, 0.0, 0.0], dtype=images.dtype, device=images.device
            )
        if images.dtype != torch.float32:
            images = to_float(images)
    if backgrounds is not None:
        if backgrounds.dtype != np.float32:
            backgrounds = to_float(backgrounds)
        bg_color = backgrounds
    alpha = images[..., 3:4]
    rgb = images[..., :3]
    return alpha * rgb + (1 - alpha) * bg_color


def draw_text_on_image(images, text_per_image, loc=(0, 0), size=48, fill_white=True):
    """
    writes text on images given as np array (b x H x W x 3)
    :param images: (b x H x W x 3) numpy array
    :param text_per_image: list or np array of strings
    :param loc: a tuple xy of anchor coordinates for the text (anchor is left-top of text)
    :param size: size of font
    :param fill_white: if True, text is white, otherwise black
    :return: new (b x H x W x 3) numpy array with text written
    """
    is_numpy = is_np(images)
    if not is_numpy:
        device = images.device
        images = to_np(images)
    is_float = images.dtype == np.float32
    if is_float:
        images = to_8b(images)
    rgbs = [Image.fromarray(x) for x in images]
    resource_path = Path(__file__).parent.resolve()
    font = ImageFont.truetype(Path(resource_path, "FreeMono.ttf"), size)
    if fill_white:
        fill = "white"
    else:
        fill = "black"
    for i, rgb in enumerate(rgbs):
        text = text_per_image[i]
        ImageDraw.Draw(rgb).text(loc, text, fill=fill, font=font)
    rgbs = np.array([np.asarray(rgb) for rgb in rgbs])
    if is_float:
        rgbs = to_float(rgbs)
    if not is_numpy:
        rgbs = to_torch(rgbs, device=device)
    return rgbs


def draw_gizmo_on_image(np_images, w2c, opengl=False, scale=0.05):
    """
    adds a gizmo to a batch of np images.
    note: will broadcast np_images and w2c against eachother.
    :param np_images: b x H x W x 3
    :param w2c: b x 3 x 4 w2c transforms (opencv conventions)
    :param opengl: if True, the w2c transforms are assumed to be in OpenGL conventions, else OpenCV conventions
    for opengl, w2c should be a bx4x4 matrix converting from world to *CLIP* space.
    :param scale: scale of the gizmo
    :return: b x H x W x 3
    """
    new_images = []
    if np_images.ndim != 4:
        raise ValueError("np_images must be b x H x W x 3")
    if w2c.ndim != 3:
        raise ValueError("KRt must be b x 3 x 4")
    np_images, w2c = broadcast_batch(np_images, w2c)
    for i, np_image in enumerate(np_images):
        pil_image = Image.fromarray(to_8b(np_image))
        W, H = pil_image.size
        gizmo_cords, _, _ = get_gizmo_coords(scale)
        gizmo_hom = to_hom(gizmo_cords)
        verts_clip = (w2c[i] @ gizmo_hom.T).T
        verts_clip = homogenize(verts_clip)
        if opengl:
            verts_clip = verts_clip[:, :2]
            verts_screen = np.array([W, H]) * (verts_clip + 1) / 2
            verts_screen[:, 1] *= -1
        else:
            verts_screen = verts_clip
        desired_loc = np.array([W - 40, H - 40])
        verts_screen += desired_loc - verts_screen[0]
        draw = ImageDraw.Draw(pil_image)
        draw.line((tuple(verts_screen[0]), tuple(verts_screen[1])), fill="red", width=0)
        draw.line(
            (tuple(verts_screen[0]), tuple(verts_screen[2])), fill="green", width=0
        )
        draw.line(
            (tuple(verts_screen[0]), tuple(verts_screen[3])), fill="blue", width=0
        )
        new_images.append(np.array(pil_image))
    return (np.array(new_images) / 255.0).astype(np.float32)


def merge_figures_with_line(
    img1,
    img2,
    lower_intersection=0.6,
    angle=np.pi / 2,
    line_width=5,
    line_color=[255, 255, 255, 255],
):
    """
    merges two np images (H x W x 3) with a white line in between
    :param img1: (H x W x 3) numpy array
    :param img2: (H x W x 3) numpy array
    :param lower_intersection: lower intersection of the line with the images
    :param angle: angle of the line
    :param line_width: width of the line
    :param line_color: color of the line
    :return: new (H x W x 3) numpy array with line in between
    """
    if img1.dtype != np.uint8:
        line_color = np.array(line_color) / 255
    else:
        line_color = np.array(line_color, dtype=np.uint8)
    combined = np.ascontiguousarray(np.broadcast_to(line_color, img1.shape))
    y, x, _ = img1.shape
    yy, xx = np.mgrid[:y, :x]
    img1_positions = (xx - lower_intersection * x) * np.tan(angle) - line_width // 2 > (
        yy - y
    )
    img2_positions = (xx - lower_intersection * x) * np.tan(angle) + line_width // 2 < (
        yy - y
    )
    combined[img1_positions] = img1[img1_positions]
    combined[img2_positions] = img2[img2_positions]
    return combined


def generate_voronoi_diagram(height, width, num_cells=1000, bg_color="white", dst=None):
    """
    generate a voronoi diagram HxWx3 with random colored cells
    :param height: height of the image
    :param width: width of the image
    :param num_cells: number of cells
    :param bg_color: background color
    :param dst: if not None, the image is written to this path
    :return: (H x W x 3) numpy array
    """
    nx = np.random.rand(num_cells) * width
    ny = np.random.rand(num_cells) * height
    nxy = np.stack((nx, ny), axis=-1)
    img = Image.new("RGB", (width, height), bg_color)
    vor = spatial.Voronoi(nxy)
    polys = vor.regions
    vertices = vor.vertices
    for poly in polys:
        polygon = vertices[poly]
        if len(poly) > 0 and np.all(np.array(poly) > 0):
            img1 = ImageDraw.Draw(img)
            img1.polygon(
                list(map(tuple, polygon)),
                fill=tuple(np.random.randint(0, 255, size=(3,))),
            )
    if dst is not None:
        img.save(str(dst))
    return np.array(img)


def generate_dot_pattern(
    height, width, background="black", radius=5, spacing=50, dst=None
):
    """
    generates an image with colored circles in a grid
    :param height: height of the image
    :param width: width of the image
    :param background: background color
    :param radius: radius of the circles in pixels
    :param spacing: spacing between the circles in pixels
    :param dst: if not None, the image is written to this path
    :return: (H x W x 3) numpy array (uint8)
    """
    img = Image.new("RGB", (width, height), background)
    img1 = ImageDraw.Draw(img)
    for i in range(0, width, spacing):
        for j in range(0, height, spacing):
            img1.ellipse(
                [i, j, i + radius * 2, j + radius * 2],
                fill=tuple(np.random.randint(0, 255, size=(3,))),
            )
    if dst is not None:
        img.save(str(dst))
    return np.array(img)


def generate_random_block_mask(size, block_size, batch_size=1):
    """
    Generates a random binary mask matrix with the specified size and block size.

    :param: size (int): Size of the square matrix (must be a power of 2).
    :param: block_size (int): Size of the square blocks (must be a power of 2 and ≤ size).

    :return: a binary np array of shape (size, size).
    """
    if not (size & (size - 1) == 0 and block_size & (block_size - 1) == 0):
        raise ValueError("Size and block_size must be powers of 2.")
    if block_size < 1:
        raise ValueError("Block size must be larger or equal to 1.")
    if block_size > size:
        raise ValueError("Block size must be less than or equal to the size.")

    num_blocks = size // block_size
    if batch_size > 1:
        random_blocks = np.random.randint(2, size=(batch_size, num_blocks, num_blocks))
        mask = np.kron(
            random_blocks, np.ones((block_size, block_size), dtype=bool)
        ).astype(bool)
    else:
        random_blocks = np.random.randint(2, size=(num_blocks, num_blocks))
        mask = np.kron(
            random_blocks, np.ones((block_size, block_size), dtype=bool)
        ).astype(bool)
    return mask


def generate_checkerboard(h, w, blocksize):
    """
    generates a checkerboard pattern
    note: if blocksize is not a divisor of w or h, the pattern will have extra "crops" at the edges
    :param h: height of the image
    :param w: width of the image
    :param blocksize: size of the squares
    :return: (H x W x 1) numpy array (bool)
    """
    c0, c1 = 0, 1  # color of the squares, for binary these are just 0,1
    tile = (
        np.array([[c0, c1], [c1, c0]], dtype=bool)
        .repeat(blocksize, axis=0)
        .repeat(blocksize, axis=1)[..., None]
    )
    grid = np.tile(tile, (h // (2 * blocksize) + 1, w // (2 * blocksize) + 1, 1))
    return grid[:h, :w]


def generate_stripe_pattern(
    height,
    width,
    background="black",
    direction="vert",
    thickness=5,
    spacing=50,
    dst=None,
):
    """
    generates an image with colored stripes in a certain direction
    :param height: height of the image
    :param width: width of the image
    :param background: background color
    :param direction: direction of the stripes ("vert", "hor", "both")
    :param thickness: thickness of the stripes
    :param spacing: spacing between the stripes
    :param dst: if not None, the image is written to this path
    :return: (H x W x 3) numpy array (uint8)
    """
    img = Image.new("RGB", (width, height), background)
    img1 = ImageDraw.Draw(img)
    if direction == "vert":
        for i in range(0, width, spacing):
            img1.rectangle(
                [i, 0, i + thickness, height],
                fill=tuple(np.random.randint(0, 255, size=(3,))),
            )
    elif direction == "hor":
        for i in range(0, height, spacing):
            img1.rectangle(
                [0, i, width, i + thickness],
                fill=tuple(np.random.randint(0, 255, size=(3,))),
            )
    elif direction == "both":
        for i in range(0, width, spacing):
            img1.rectangle(
                [i, 0, i + thickness, height],
                fill=tuple(np.random.randint(0, 255, size=(3,))),
            )
        for i in range(0, height, spacing):
            img1.rectangle(
                [0, i, width, i + thickness],
                fill=tuple(np.random.randint(0, 255, size=(3,))),
            )
    else:
        raise ValueError("direction must be either 'vert', 'hor' or 'both'")
    if dst is not None:
        img.save(str(dst))
    return np.array(img)


def generate_gaussian_image(height, width, center=(0, 0), sigma=(10, 10), theta=0):
    """
    generate an image of a gaussian
    :param: height: the height of the image
    :param: width: the width of the image
    :param: center: the x,y center of the gaussian (default is upper left corner)
    :param: theta: rotation of the gaussian
    :param: sigma: the sx,sy stdev in the x and y acis before rotation
    :return: the gaussian image
    """
    theta = 2 * np.pi * theta / 360
    x = np.arange(0, width, 1, np.float32)
    y = np.arange(0, height, 1, np.float32)
    y = y[:, np.newaxis]
    sx = sigma[0]
    sy = sigma[1]
    x0 = center[0]
    y0 = center[1]

    # rotation
    a = np.cos(theta) * x - np.sin(theta) * y
    b = np.sin(theta) * x + np.cos(theta) * y
    a0 = np.cos(theta) * x0 - np.sin(theta) * y0
    b0 = np.sin(theta) * x0 + np.cos(theta) * y0

    return np.exp(-(((a - a0) ** 2) / (2 * (sx**2)) + ((b - b0) ** 2) / (2 * (sy**2))))


def generate_lollipop_pattern(height, width, background="black", n=15, m=8, dst=None):
    """
    generates an image with a lollipop pattern
    :param height: height of the image
    :param width: width of the image
    :param background: background color
    :param n: number of circles in the pattern
    :param m: number of lines in the pattern
    :param dst: if not None, the image is written to this path
    :return: (H x W x 3) numpy array (uint8)
    """
    spacing_x = width // (2 * n)
    spacing_y = height // (2 * n)
    spacing_angle = 360 // m
    img = Image.new("RGB", (width, height), background)
    img1 = ImageDraw.Draw(img)
    for j in range(m):
        for i in range(n):
            x0 = spacing_x * i
            y0 = spacing_y * i
            start_angle = j * spacing_angle
            end_angle = 360
            img1.pieslice(
                [x0, y0, width - x0, height - y0],
                start=start_angle,
                end=end_angle,
                fill=tuple(np.random.randint(0, 255, size=(3,))),
            )
    if dst is not None:
        img.save(str(dst))
    return np.array(img)


def generate_concentric_circles(
    height, width, background="black", n=5, colors=None, dst=None
):
    """
    generates an image with colored concentric circles
    :param height: height of the image
    :param width: width of the image
    :param background: background color (as int, tuple of ints or name accepted by PIL)
    :param n: number of circles to draw
    :param colors: if not None, list of n colors of the circles where each color is a list of 3 \in [0,255]
    :param dst: if not None, the image is written to this path
    """
    spacing_x = width // (2 * n)
    spacing_y = height // (2 * n)
    img = Image.new("RGB", (width, height), background)
    img1 = ImageDraw.Draw(img)
    for i in range(n):
        x0 = spacing_x * i
        y0 = spacing_y * i
        if colors is None:
            img1.ellipse(
                [x0, y0, width - x0, height - y0],
                fill=tuple(np.random.randint(0, 255, size=(3,))),
            )
        else:
            img1.ellipse(
                [x0, y0, width - x0, height - y0],
                fill=tuple(colors[i]),
            )
    if dst is not None:
        img.save(str(dst))
    return np.array(img)


def generate_gray_gradient(
    height, width, grayscale=False, vertical=True, flip=False, bins=10
):
    """
    generate a gray gradient image HxWx3
    :param height: height of the image
    :param width: width of the image
    :param grayscale: if True, the image is grayscale
    :param vertical: if True, the gradient is vertical
    :param flip: if True, the gradient is flipped
    :param bins: number of bins
    :return: (H x W x 3) uint8 numpy array
    """
    bins = np.clip(bins, 1, 256)
    colors = np.linspace(0, 255, num=bins).astype(np.uint8)
    if vertical:
        n_bins = height // bins
        up_to_max_intensity = colors.repeat(n_bins)[:height]
        pad_size = height - up_to_max_intensity.shape[0]
        if pad_size > 0:
            channel = np.concatenate(
                (up_to_max_intensity, np.full(pad_size, 255, dtype=np.uint8))
            )
        else:
            channel = up_to_max_intensity
        img = channel[:, None].repeat(width, axis=1)
        if flip:
            img = np.flip(img, axis=0)
    else:
        n_bins = width // bins
        up_to_max_intensity = colors.repeat(n_bins)[:width]
        pad_size = width - up_to_max_intensity.shape[0]
        if pad_size > 0:
            channel = np.concatenate(
                (up_to_max_intensity, np.full(pad_size, 255, dtype=np.uint8))
            )
        else:
            channel = up_to_max_intensity
        img = channel[None, :].repeat(height, axis=0)
        if flip:
            img = np.flip(img, axis=1)
    if not grayscale:
        img = img[:, :, None].repeat(3, axis=-1)
    return img


def image_grid(images, rows, cols, pad=0, pad_color=None):
    """
    :param images: list of images
    :param rows: number of rows
    :param cols: number of cols
    :param pad: will pad images by this number of pixels with pad_color
    :param pad_color: a (3,) np array representing the pad color. if not provided pad will be black. must be same dtype as images.
    :return: grid image
    """
    if images.ndim != 4:
        raise ValueError("images must be a 4D array")
    if len(images) != rows * cols:
        raise ValueError("number of images must be equal to rows * cols")
    if pad > 0:
        images = pad_to_res(
            images, images.shape[1] + pad * 2, images.shape[2] + pad * 2, pad_color
        )
    tmp = images.reshape(rows, cols, images.shape[1], images.shape[2], -1)
    if type(tmp) == torch.Tensor:
        result = tmp.permute(0, 2, 1, 3, 4).reshape(
            rows * images.shape[1], cols * images.shape[2], -1
        )
    elif type(tmp) == np.ndarray:
        result = tmp.transpose(0, 2, 1, 3, 4).reshape(
            rows * images.shape[1], cols * images.shape[2], -1
        )
    return result


def resize(images, H, W, mode="bilinear"):
    """
    wrapper around torch interpolate (https://pytorch.org/docs/stable/generated/torch.nn.functional.interpolate.html)
    :param images: batch of np array (b, h, w, c) or torch tensors (b, c, h, w)
    :param H: output height
    :param W: output width
    :param mode: pass through for torch ()
    :return: same as input, with the new H and W
    """
    if images.ndim != 4:
        raise ValueError("images must be a 4D array")
    was_numpy = False
    if is_np(images):
        imgs_torch = to_torch(images).permute(2, 0, 1)
        was_numpy = True
    else:
        imgs_torch = images
    interpolated = torch.nn.functional.interpolate(
        imgs_torch,
        size=(H, W),
        scale_factor=None,
        mode=mode,
        align_corners=None,
        recompute_scale_factor=None,
        antialias=False,
    )
    if was_numpy:
        interpolated = to_np(interpolated.permute(1, 2, 0))
    return interpolated


def resize_images_naive(images, H, W, channels_last=True, mode="mean"):
    """
    resize images to output_size, but only if the output size has a common divisor of the input size
    :param images: numpy array of images (N x H x W x C)
    :param H: output height size that has a common divisor with the input height size
    :param W: output width size that has a common divisor with the input width size
    :param channels_last: if True, the images are provided in channels last format (and so will the output)
    :param mode: one of "max", "mean"
    :return: np array of resized images
    """
    if images.ndim != 4:
        raise ValueError("images must be a 4D array")
    if images.shape[1] != images.shape[2]:
        raise ValueError("images must be square")
    if not channels_last:
        raise NotImplementedError("only channels last is supported")
    channels_size = images.shape[-1]
    input_size = np.array(images.shape[1:3])
    output_size = np.array([H, W])
    bin_size = input_size // output_size
    if mode == "max":
        small_images = (
            images.reshape(
                (
                    images.shape[0],
                    output_size[0],
                    bin_size[0],
                    output_size[1],
                    bin_size[1],
                    channels_size,
                )
            )
            .max(4)
            .max(2)
        )
    elif mode == "mean":
        small_images = (
            images.reshape(
                (
                    images.shape[0],
                    output_size[0],
                    bin_size[0],
                    output_size[1],
                    bin_size[1],
                    channels_size,
                )
            )
            .mean(4)
            .mean(2)
        )
        if images.dtype == np.uint8:
            small_images = small_images.astype(np.uint8)
    else:
        raise ValueError("mode must be one of 'max', 'mean'")
    return small_images


def pad_to_square(images, color=None):
    """
    pads a batch of images to a square shape
    note: smaller dimension is padded
    :param image: numpy image b x h x w x c
    :param color: color to pad with
    :return: padded image
    """
    if images.ndim != 4:
        raise ValueError("image must be a 4D array")
    diff = images.shape[1] - images.shape[2]
    if diff == 0:
        return images
    if diff > 0:
        return pad_to_res(images, images.shape[1], images.shape[1], color)
    else:
        return pad_to_res(images, images.shape[2], images.shape[2], color)


def crop_to_square(images):
    """
    crops a batch of images to a square shape
    note: bigger dimension is cropped
    :param img: numpy image h x w x c
    :return: the cropped square image
    """
    if images.ndim != 4:
        raise ValueError("image must be a 4D array")
    if images.shape[1] > images.shape[2]:
        s = int((images.shape[1] - images.shape[2]) / 2)
        return images[s : (s + images.shape[2])]
    else:
        s = int((images.shape[2] - images.shape[1]) / 2)
        return images[:, :, s : (s + images.shape[1])]


def pad_to_res(images, res_h, res_w, bg_color=None):
    """
    pads a batch of numpy images to a specific resolution
    :param image: numpy image b x h x w x c
    :param res_h: height of the output image
    :param res_w: width of the output image
    :param bg_color: background color sized (c,) (defaults to black)
    :return: padded image b x res_h x res_w x c
    """
    if bg_color is None:
        if is_np(images):
            bg_color = np.zeros(images.shape[-1], dtype=images.dtype)
        else:
            bg_color = torch.zeros(
                images.shape[-1], dtype=images.dtype, device=images.device
            )
    if images.ndim != 4:
        raise ValueError("image must be a 4D array")
    b, h, w, c = images.shape
    if h > res_h or w > res_w:
        raise ValueError("images dimensions is larger than the output resolution")
    if h == res_h and w == res_w:
        return images
    if bg_color.shape[0] != c:
        raise ValueError(
            "background color must have the same number of channels as the image"
        )
    bg_color = bg_color[None, None, None, :]
    if is_np(images):
        output = np.zeros((b, res_h, res_w, c), dtype=images.dtype)
    else:
        output = torch.zeros(
            (b, res_h, res_w, c), dtype=images.dtype, device=images.device
        )
    output[:, :, :, :] = bg_color
    corner_left = (res_w - w) // 2
    corner_top = (res_h - h) // 2
    output[:, corner_top : corner_top + h, corner_left : corner_left + w, :] = images
    return output


def crop_center(images, dst_h, dst_w):
    """
    crops a batch of images to a specific resolution, but the crop comes from the center of the image
    :param image: numpy (or torch) array b x h x w x c
    :param dst_h: height of the output image
    :param dst_w: width of the output image
    :return: cropped image b x dst_h x dst_w x c
    """
    if images.ndim != 4:
        raise ValueError("image must be a 4D array")
    _, h, w, _ = images.shape
    if h < dst_h or w < dst_w:
        raise ValueError("images dimensions is smaller than the output resolution")
    if h == dst_h and w == dst_w:
        return images
    corner_left = (w - dst_w) // 2
    corner_top = (h - dst_h) // 2
    output = images[
        :, corner_top : corner_top + dst_h, corner_left : corner_left + dst_w, :
    ]
    return output


def mask_regions(images, start_h, end_h, start_w, end_w):
    """
    masks a batch of numpy image with black background outside of region of interest (roi)
    :param image: numpy image b x h x w x c
    :param start_h: where does the roi height start
    :param end_h: where does the roi height end
    :param start_w: where does the roi width start
    :param end_w: where does the roi width end
    :return: masked image
    """
    if images.ndim != 4:
        raise ValueError("image must be a 4D array")
    _, h, w, _ = images.shape
    if start_h < 0 or start_w < 0 or end_h > h or end_w > w:
        raise ValueError("Values exceed image resolution")
    output = np.zeros_like(images)
    output[:, start_h:end_h, start_w:end_w, :] = images[start_h:end_h, start_w:end_w, :]
    return output


def adjust_contrast_brightness(img, alpha, beta=None):
    """
    adjusts image contrast and brightness using naive gain and bias factors
    :param img: input image numpy array (n x h x w x 3), float values between 0 and 1
    :param alpha: gain factor ("contrast") between 0 and inf
    :param beta: bias factor ("brightness") between -inf and inf (but typically between -1 and 1)
    :return: the new image
    """
    if img.dtype != np.float32:
        raise ValueError("img must be a float32 numpy array (0-1)")
    if beta is None:  # if beta is not provided, set to factor of alpha
        beta = 0.5 - alpha / 2
    new_img = img * alpha + beta
    new_img[new_img < 0] = 0
    new_img[new_img > 1] = 1
    return new_img.astype(np.uint8)


def change_brightness(input_img, brightness=0):
    """
    changes brightness of an image or batch of images
    :param input_img a numpy or torch tensor of float values between 0 and 1 (n x h x w x 3)
    :param brightness a number between -255 to 255 (0=no change)
    :return the new image
    """
    if input_img.dtype != np.float32 and input_img.dtype != torch.float32:
        raise ValueError("input_img must be a float32 array (0-1)")
    if brightness != 0:
        if brightness > 0:
            shadow = brightness
            highlight = 255
        else:
            shadow = 0
            highlight = 255 + brightness
        alpha_b = (highlight - shadow) / 255
        gamma_b = shadow / 255
    else:
        alpha_b = 1
        gamma_b = 0
    return input_img * alpha_b + gamma_b


def linear_to_srgb(linear):
    """
    converts linear RGB to sRGB, see https://en.wikipedia.org/wiki/SRGB.
    note: linear is expected to be in the range [0, 1]
    """
    eps = np.finfo(np.float32).eps
    srgb0 = 323 / 25 * linear
    srgb1 = (211 * np.maximum(eps, linear) ** (5 / 12) - 11) / 200
    return np.where(linear <= 0.0031308, srgb0, srgb1)


def srgb_to_linear(srgb):
    """
    converts linear RGB to sRGB, see https://en.wikipedia.org/wiki/SRGB.
    note: srgb is expected to be in the range [0, 1]
    """
    eps = np.finfo(np.float32).eps
    linear0 = 25 / 323 * srgb
    linear1 = np.maximum(eps, ((200 * srgb + 11) / (211))) ** (12 / 5)
    return np.where(srgb <= 0.04045, linear0, linear1)


def compute_color_distance(image1, image2, bin_per_dim=10):
    """
    computes a naive "color distance" between two images by binning the colors and computing the wasserstein distance per channel
    :param image1: numpy image h x w x 3
    :param image2: numpy image h x w x 3
    :return: the sum (over channels) of the wasserstein distances
    """
    if image1.shape != image2.shape:
        raise ValueError("images must have the same shape")
    if image1.ndim != 3:
        raise ValueError("images must be 3D arrays")
    if image1.dtype != np.uint8:
        raise ValueError("images must be uint8")
    spatial_size = image1.shape[0] * image1.shape[1]
    hist1_R, _ = np.histogram(image1[:, :, 0].reshape(-1), bins=bin_per_dim)
    hist2_R, _ = np.histogram(image2[:, :, 0].reshape(-1), bins=bin_per_dim)
    hist1_R = hist1_R / spatial_size
    hist2_R = hist2_R / spatial_size
    result_R = wasserstein_distance(hist1_R, hist2_R)
    ###
    hist1_G, _ = np.histogram(image1[:, :, 1].reshape(-1), bins=bin_per_dim)
    hist2_G, _ = np.histogram(image2[:, :, 1].reshape(-1), bins=bin_per_dim)
    hist1_G = hist1_G / spatial_size
    hist2_G = hist2_G / spatial_size
    result_G = wasserstein_distance(hist1_G, hist2_G)
    ###
    hist1_B, _ = np.histogram(image1[:, :, 2].reshape(-1), bins=bin_per_dim)
    hist2_B, _ = np.histogram(image2[:, :, 2].reshape(-1), bins=bin_per_dim)
    hist1_B = hist1_B / spatial_size
    hist2_B = hist2_B / spatial_size
    result_B = wasserstein_distance(hist1_B, hist2_B)
    result = result_R + result_G + result_B
    return result


def tonemap(
    hdr_image, exposure=0.0, offset=0.0, gamma=2.2, only_preproc=False, clip=True
):
    """
    maps an input image [-inf, inf] to [0, 1] using non-linear gamma correction.
    this slightly naive tonemapping was taken from https://github.com/Tom94/tev
    :param hdr_image: a numpy array or torch tensor (channels first or last, any float type)
    :param exposure: the image will be multiplied by 2*exposure prior to gamma correction
    :param offset: will be added to image (after multiplied by exposure, but prior to gamma correction)
    :param gamma: gamma to be used for non-linear correction
    :param only_preproc: if true, will not run gamma correction but only use exposure and offset
    :param clip: if true will clip result to [0.0, 1.0]
    :return: the tonemapped image, with same dtype and shape
    """
    if type(hdr_image) == np.ndarray:
        image = (hdr_image * np.power(2.0, exposure)) + offset
        if not only_preproc:
            image = np.sign(image) * np.power(np.abs(image), 1.0 / gamma)
        if clip:
            image = np.clip(image, 0.0, 1.0)
    elif type(hdr_image) == torch.Tensor:
        image = (hdr_image * np.power(2.0, exposure)) + offset
        if not only_preproc:
            image = torch.sign(image) * torch.pow(torch.abs(image), 1.0 / gamma)
        if clip:
            image = torch.clamp(image, 0.0, 1.0)
    else:
        raise TypeError("hdr_image must be either a numpy array or torch tensor")
    return image



import pytest
import numpy as np
import torch
import gsoup
from pathlib import Path
from gsoup.viewer_drivers import calibration_static_view


def test_type_conversions():
    test_numpy_bool = np.array([1, 0, 0, 1], dtype=bool)
    test_torch_bool = gsoup.to_torch(test_numpy_bool)

    test_float_from_bool = gsoup.to_float(test_numpy_bool)
    assert test_float_from_bool.dtype == np.float32
    test_8bit_from_bool = gsoup.to_8b(test_numpy_bool)
    assert test_8bit_from_bool.dtype == np.uint8
    test_8bit_from_float = gsoup.to_8b(test_float_from_bool)
    assert test_8bit_from_float.dtype == np.uint8
    test_float_from_8bit = gsoup.to_float(test_8bit_from_float)
    assert test_float_from_8bit.dtype == np.float32
    test_float_from_bool = gsoup.to_float(test_torch_bool)
    assert test_float_from_bool.dtype == torch.float32
    test_8bit_from_bool = gsoup.to_8b(test_torch_bool)
    assert test_8bit_from_bool.dtype == torch.uint8
    test_8bit_from_float = gsoup.to_8b(test_float_from_bool)
    assert test_8bit_from_float.dtype == torch.uint8
    test_float_from_8bit = gsoup.to_float(test_8bit_from_float)
    assert test_float_from_8bit.dtype == torch.float32


def test_broadcast_batch():
    R = np.random.randn(1, 3, 3)
    t = np.random.randn(1, 3)
    R, t = gsoup.broadcast_batch(R, t)
    assert R.shape == (1, 3, 3)
    assert t.shape == (1, 3)
    R = np.random.randn(2, 3, 3)
    t = np.random.randn(2, 3)
    R, t = gsoup.broadcast_batch(R, t)
    assert R.shape == (2, 3, 3)
    assert t.shape == (2, 3)
    R = np.random.randn(2, 3, 3)
    t = np.random.randn(1, 3)
    R, t = gsoup.broadcast_batch(R, t)
    assert R.shape == (2, 3, 3)
    assert t.shape == (2, 3)
    R = np.random.randn(1, 3, 3)
    t = np.random.randn(2, 3)
    R, t = gsoup.broadcast_batch(R, t)
    assert R.shape == (2, 3, 3)
    assert t.shape == (2, 3)
    R = np.random.randn(3)
    t = np.random.randn(3)
    R, t = gsoup.broadcast_batch(R, t)
    assert R.shape == (1, 3)
    assert t.shape == (1, 3)
    R = np.random.randn(1)
    t = np.random.randn(1)
    R, t = gsoup.broadcast_batch(R, t)
    assert R.shape == (1, 1)
    assert t.shape == (1, 1)
    R = np.random.randn(3, 3)
    t = np.random.randn(3)
    with pytest.raises(ValueError):
        R, t = gsoup.broadcast_batch(R, t)


def test_transforms():
    R = np.random.randn(2, 3, 3)
    t = np.random.randn(1, 3)
    Rt = gsoup.compose_rt(R, t)
    assert Rt.shape == (2, 3, 4)

    eye = np.array([1, 0, 0], dtype=np.float32)
    at = np.array([0, 0, 0], dtype=np.float32)
    up = np.array([0, 0, 1], dtype=np.float32)
    transform_np = gsoup.look_at_np(eye, at, up)
    transform_torch = gsoup.look_at_torch(
        gsoup.to_torch(eye), gsoup.to_torch(at), gsoup.to_torch(up)
    )
    assert np.allclose(transform_np, gsoup.to_numpy(transform_torch))
    transform_np_opengl = gsoup.look_at_np(eye, at, up, opengl=True)
    transform_torch_opengl = gsoup.look_at_torch(
        gsoup.to_torch(eye), gsoup.to_torch(at), gsoup.to_torch(up), opengl=True
    )
    assert np.allclose(transform_np_opengl, gsoup.to_numpy(transform_torch_opengl))
    normal = torch.tensor([0, 0, 1.0])

    location3d = np.array([0.1, 0.1, 0.1])
    location3d_noise = location3d + np.random.randn(3) * 0.01
    location4d = gsoup.to_hom(location3d)
    location4d_noise = gsoup.to_hom(location3d_noise)
    # sanity on opengl projection
    v2w = gsoup.look_at_np(
        np.array([1.0, 0, 0]), location3d, np.array([0, 0, 1.0]), opengl=True
    )[0]
    w2v = np.linalg.inv(v2w)
    v2c = gsoup.perspective_projection()
    opengl_location = v2c @ w2v @ location4d
    opengl_location = gsoup.homogenize(opengl_location)
    assert np.allclose(opengl_location[:2], np.zeros(2))
    # sanity on opencv projection
    c2w = gsoup.look_at_np(np.array([1.0, 0, 0]), location3d, np.array([0, 0, 1.0]))[0]
    w2c = np.linalg.inv(c2w)
    K = gsoup.opencv_intrinsics_from_opengl_project(v2c, 1, 1)
    opencv_location = K @ gsoup.to_34(w2c) @ location4d
    opencv_location = gsoup.homogenize(opencv_location)
    assert np.allclose(opencv_location[:2], np.ones(2) * 0.5)
    # test opencv / opengl conversion
    opengl_location = v2c @ w2v @ location4d_noise
    opengl_location = gsoup.homogenize(opengl_location)
    opencv_location = K @ gsoup.to_34(w2c) @ location4d_noise
    opencv_location = gsoup.homogenize(opencv_location)
    # x is in same direction, but opengl screen is -1 to 1 (while opencv is 0 to 1)
    assert np.allclose((opencv_location[0] - opengl_location[0] / 2), 0.5)
    # y is in opposite direction, but opengl screen is -1 to 1 (while opencv is 0 to 1)
    assert np.allclose((opencv_location[1] + opengl_location[1] / 2), 0.5)


def test_rotations():
    qvecs = gsoup.random_qvec(10)
    torch_qvecs = torch.tensor(qvecs)
    rotmats = gsoup.qvec2mat(qvecs)
    assert rotmats.shape == (10, 3, 3)
    rotmats = gsoup.qvec2mat(torch_qvecs)
    assert rotmats.shape == (10, 3, 3)
    new_qvecs = gsoup.mat2qvec(rotmats)
    mask1 = torch.abs(new_qvecs - torch_qvecs) < 1e-6
    mask2 = torch.abs(new_qvecs + torch_qvecs) < 1e-6
    assert torch.all(mask1 | mask2)
    rotmat = gsoup.qvec2mat(torch_qvecs[0:1])
    assert rotmat.shape == (1, 3, 3)
    new_qvec = gsoup.mat2qvec(rotmat)
    mask1 = torch.abs(new_qvec[0] - torch_qvecs[0]) < 1e-6
    mask2 = torch.abs(new_qvec[0] + torch_qvecs[0]) < 1e-6
    assert torch.all(mask1 | mask2)
    normal = torch.tensor([0, 0, 1.0])
    random_vectors = gsoup.random_vectors_on_sphere(10, normal=normal)
    assert random_vectors.shape == (10, 3)
    assert (random_vectors @ normal).all() > 0
    normal = torch.tensor([[0, 0, 1.0]]).repeat(10, 1)
    random_vectors = gsoup.random_vectors_on_sphere(10, normal=normal)
    assert random_vectors.shape == (10, 3)
    assert (random_vectors[:, None, :] @ normal[:, :, None]).all() > 0
    rotx = gsoup.rotx(np.pi / 2, degrees=False)
    assert np.allclose(
        rotx, np.array([[1.0, 0, 0, 0], [0, 0, -1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])
    )


def test_homogenize():
    x = np.random.rand(100, 2)
    hom_x = gsoup.to_hom(x)
    assert hom_x.shape == (100, 3)
    hom_x = gsoup.to_hom(hom_x)
    assert hom_x.shape == (100, 4)
    dehom_x = gsoup.homogenize(hom_x)
    assert dehom_x.shape == (100, 3)
    dehom_x = gsoup.homogenize(dehom_x, keepdim=True)
    assert dehom_x.shape == (100, 3)
    x = np.random.rand(3)
    hom_x = gsoup.to_hom(x)
    assert hom_x.shape == (4,)
    dehom_x = gsoup.homogenize(hom_x)
    assert dehom_x.shape == (3,)


def test_normalize_vertices():
    v = np.random.rand(100, 3) * 100
    v_normalized = gsoup.normalize_vertices(v)
    assert (v_normalized < 1.0).all()

    v = torch.rand(100, 3) * 100
    v_normalized = gsoup.normalize_vertices(v)
    assert (v_normalized < 1.0).all()


def test_point_to_line_distance():
    p = np.array([0, 0, 0])
    v0 = np.array([1, 0, 0])
    v1 = np.array([2, 0, 0])
    dist = gsoup.point_line_distance(p, v0, v1)
    assert dist == 0.0
    p = np.array([0, 1, 0])
    dist = gsoup.point_line_distance(p, v0, v1)
    assert dist == 1.0
    v0 = np.array([0, 1, 0])
    v1 = np.array([0, 2, 0])
    p = np.array([0, 0, 0])
    dist = gsoup.point_line_distance(p, v0, v1)
    assert dist == 0.0
    p = np.array([1, 0, 0])
    dist = gsoup.point_line_distance(p, v0, v1)
    assert dist == 1.0
    p = np.array([[0, 0, 1], [1, 0, 0]])
    dist = gsoup.point_line_distance(p, v0, v1)
    assert (dist == np.array([1.0, 1.0])).all()


def test_structures():
    v, f = gsoup.structures.cube()
    assert v.shape[0] == 8
    assert f.shape[0] == 12
    v, f = gsoup.structures.icosehedron()
    assert v.shape[0] == 12
    assert f.shape[0] == 20


def test_parsers():
    # obj save and load
    v, f = gsoup.structures.quad_cube()
    gsoup.save_mesh(v, f, "resource/quad_cube.obj")
    gsoup.save_mesh(v, f, "resource/quad_cube.ply")
    v, f = gsoup.structures.cube()
    gsoup.save_mesh(v, f, "resource/cube.obj")
    v1, f1 = gsoup.load_mesh("resource/cube.obj")
    assert np.allclose(v, v1)
    assert np.allclose(f, f1)
    v, f = gsoup.structures.icosehedron()
    vc = np.random.randint(0, 255, size=v.shape).astype(np.uint8)
    gsoup.save_mesh(v, f, "resource/ico.obj")
    v1, f1 = gsoup.load_mesh("resource/ico.obj")
    assert np.allclose(v, v1)
    assert np.allclose(f, f1)
    # ply save
    gsoup.save_mesh(v, f, "resource/ico.ply")
    gsoup.save_mesh(v, f, "resource/ico_vcolor.ply", vertex_colors=vc)
    gsoup.save_mesh(
        v,
        f,
        "resource/ico_fcolor.ply",
        face_colors=np.random.randint(0, 255, size=f.shape).astype(np.uint8),
    )
    gsoup.save_pointcloud(v, "resource/ico_pc.ply")
    # test ascii ply loader
    v1, f1, vc1 = gsoup.load_mesh("resource/ico_vcolor.ply", return_vert_color=True)
    assert np.allclose(v, v1)
    assert np.allclose(f, f1)
    assert np.allclose(vc, vc1)
    # more use cases
    v1, f1 = gsoup.load_mesh("resource/ico_pc.ply")  # loads a pointcloud as a mesh
    assert np.allclose(v, v1)
    assert f1 is None
    v1 = gsoup.load_pointcloud("resource/ico_pc.ply")
    assert np.allclose(v, v1)
    # test binary ply loader``
    v, channels = gsoup.load_pointcloud(
        "tests/tests_resource/splat.ply", return_vert_norms=True
    )
    assert channels.shape == (v.shape[0], 59)


def test_exr():
    normals = gsoup.read_exr("tests/tests_resource/normal0001.exr")
    gsoup.write_exr(normals, "resource/normals.exr")


def test_image():
    random_mask = gsoup.generate_random_block_mask(4, 2, 3)
    assert random_mask.shape == (3, 4, 4)
    random_mask = gsoup.generate_random_block_mask(4, 2, 1)
    assert random_mask.shape == (4, 4)
    checkboard = gsoup.generate_checkerboard(512, 512, 8)  # H, W, 1
    gsoup.save_image(checkboard, "resource/checkboard.png")
    gsoup.save_images([checkboard], "resource", file_names=["resource/checkboard.png"])
    checkboard_RGB = np.tile(checkboard, (1, 1, 3))  # H, W, 3
    checkboard_RGBA = gsoup.add_alpha(checkboard_RGB, checkboard)
    checkboard_RGBA2 = np.tile(checkboard, (1, 1, 4))  # H, W, 4
    assert (checkboard_RGBA == checkboard_RGBA2).all()
    checkboard_RGB_batch = np.tile(checkboard[None, ...], (10, 1, 1, 3))
    checkboard_RGBA_batch = gsoup.add_alpha(checkboard_RGB_batch, checkboard[None, ...])
    checkboard_RGB = gsoup.alpha_compose(checkboard_RGBA, ~checkboard_RGBA[..., :3])
    assert (checkboard_RGB == 1.0).all()
    checkboard_RGB = gsoup.alpha_compose(
        checkboard_RGBA, bg_color=np.array([0.0, 0.0, 1.0])
    )
    assert (checkboard_RGB[..., -1] == 1.0).all()
    lollipop_path = Path("resource/lollipop.png")
    lollipop = gsoup.generate_lollipop_pattern(512, 512, dst=lollipop_path)
    gsoup.save_image(lollipop, lollipop_path)
    lollipop2 = gsoup.load_image(lollipop_path)
    assert np.allclose(lollipop, lollipop2)
    gsoup.save_images(
        lollipop[None, ...], lollipop_path.parent, file_names=["test_save.png"]
    )
    lollipop_pad = gsoup.pad_to_res(lollipop[None, ...], 512, 1024)
    assert lollipop_pad.shape == (1, 512, 1024, 3)
    lollipop_padded_square = gsoup.pad_to_square(lollipop_pad)
    assert lollipop_padded_square.shape == (1, 1024, 1024, 3)
    lollipop_cropped_square = gsoup.crop_to_square(lollipop_pad)
    assert lollipop_cropped_square.shape == (1, 512, 512, 3)
    lollipop_srgb = gsoup.linear_to_srgb(lollipop)
    lollipop_linear = gsoup.srgb_to_linear(lollipop_srgb)
    assert np.allclose(lollipop, lollipop_linear)
    gsoup.generate_concentric_circles(256, 512, dst=Path("resource/circles.png"))
    gsoup.generate_stripe_pattern(
        256, 512, direction="both", dst=Path("resource/stripe.png")
    )
    gsoup.generate_dot_pattern(512, 256, dst=Path("resource/dots.png"))
    gray1 = gsoup.generate_gray_gradient(256, 256, grayscale=True)
    assert gray1.shape == (256, 256)
    assert len(np.unique(gray1)) == 10
    assert gray1.max() == 255
    gray2 = gsoup.generate_gray_gradient(50, 800, vertical=False)
    assert gray2.shape == (50, 800, 3)
    assert gray2.max() == 255
    gray3 = gsoup.generate_gray_gradient(256, 256, bins=-65)
    assert gray3.max() == 0
    gray4 = gsoup.generate_gray_gradient(256, 256, bins=300)
    assert gray4.max() == 255
    gray5 = gsoup.generate_gray_gradient(1080, 1920, bins=300)
    assert gray5.shape == (1080, 1920, 3)
    assert gray5.max() == 255
    dst = Path("resource/voronoi.png")
    gsoup.generate_voronoi_diagram(512, 512, 1000, dst=dst)
    img = gsoup.load_image(dst)
    assert img.shape == (512, 512, 3)
    assert img.dtype == np.uint8
    img = gsoup.load_image(dst, as_grayscale=True)
    assert img.shape == (512, 512, 1)
    assert img.dtype == np.uint8
    img = gsoup.load_image(dst, as_float=True)
    assert img.shape == (512, 512, 3)
    assert img.dtype == np.float32
    assert (img >= 0.0).all()
    assert (img <= 1.0).all()
    img = gsoup.load_image(dst, channels_last=False)
    assert img.shape == (3, 512, 512)
    img = gsoup.load_image(dst, as_float=True, as_grayscale=True)
    assert img.shape == (512, 512, 1)
    assert img.dtype == np.float32
    assert (img >= 0.0).all()
    assert (img <= 1.0).all()
    img = gsoup.load_image(dst, channels_last=False, as_grayscale=True)
    assert img.shape == (1, 512, 512)
    img = gsoup.load_images([dst])
    assert img.shape == (1, 512, 512, 3)
    img = gsoup.load_images([dst], as_grayscale=True)
    assert img.shape == (1, 512, 512, 1)
    img = gsoup.load_images([dst])
    assert img.shape == (1, 512, 512, 3)
    img = gsoup.load_images([dst, dst, dst, dst])
    assert img.shape == (4, 512, 512, 3)
    resized_img = gsoup.resize_images_naive(img, 256, 256, mode="mean")
    assert resized_img.shape == (4, 256, 256, 3)
    resized_img_float = gsoup.resize_images_naive(
        gsoup.to_float(img), 256, 256, mode="mean"
    )
    assert resized_img_float.shape == (4, 256, 256, 3)
    assert resized_img_float.dtype == np.float32
    resized_img_gray = gsoup.resize_images_naive(img[..., 0:1], 256, 256, mode="mean")
    assert resized_img_gray.shape == (4, 256, 256, 1)
    grid = gsoup.image_grid(resized_img, 2, 2)
    assert grid.shape == (512, 512, 3)
    white_images = gsoup.to_8b(np.ones((4, 256, 256, 3), dtype=np.float32))
    pad = 5
    grid2 = gsoup.image_grid(
        white_images, 2, 2, pad=5, pad_color=np.array([255, 255, 0])
    )
    assert grid2.shape == (512 + pad * 4, 512 + pad * 4, 3)
    img = gsoup.load_images([dst, dst, dst, dst], as_grayscale=True)
    assert img.shape == (4, 512, 512, 1)
    img = gsoup.load_images(
        [dst, dst, dst, dst], as_grayscale=True, channels_last=False
    )
    assert img.shape == (4, 1, 512, 512)
    img = gsoup.load_images([dst, dst, dst, dst], resize_wh=(128, 128))
    assert img.shape == (4, 128, 128, 3)
    img, paths = gsoup.load_images(
        [dst, dst, dst, dst],
        resize_wh=(128, 256),
        as_grayscale=True,
        channels_last=False,
        return_paths=True,
        as_float=True,
        to_torch=True,
    )
    assert len(paths) == 4
    assert img.dtype == torch.float32
    assert img.shape == (4, 1, 256, 128)
    lollipop_path1 = Path("resource/lp_ref.png")
    lollipop_path2 = Path("resource/lp_01.png")
    lollipop_path3 = Path("resource/lp_02.png")
    img_ref = gsoup.generate_lollipop_pattern(512, 512, dst=lollipop_path1)
    img1 = gsoup.to_8b(gsoup.to_float(img_ref) / 2)
    gsoup.save_image(img1, lollipop_path2)
    img2 = gsoup.generate_lollipop_pattern(512, 512, dst=lollipop_path3)
    dist1 = gsoup.compute_color_distance(img_ref, img1)
    dist2 = gsoup.compute_color_distance(img_ref, img2)
    assert dist1 < dist2
    test = np.zeros((1, 512, 512, 3), dtype=np.uint8)
    test = gsoup.draw_text_on_image(test, np.array(["target"]))
    assert np.any(test) > 0
    color_image = np.random.uniform(size=(512, 512, 3))
    gray_image = gsoup.color_to_gray(color_image)
    assert gray_image.shape == (512, 512, 1)
    gray_image = gsoup.color_to_gray(color_image, keep_channels=True)
    assert gray_image.shape == (512, 512, 3)
    assert np.all(gray_image[:, :, 0] == gray_image[:, :, 1])
    assert np.all(gray_image[:, :, 1] == gray_image[:, :, 2])


# def test_video():
#     import platform
#     import os

#     if platform.system() == "Windows":
#         FFMPEG_DIR = os.path.join("C:/tools/ffmpeg-7.0.1-essentials_build/bin")
#         os.environ["PATH"] = FFMPEG_DIR + ";" + os.environ["PATH"]
#     else:
#         FFMPEG_DIR = os.path.join("/usr/bin")
#         os.environ["PATH"] = FFMPEG_DIR + ":" + os.environ["PATH"]
#     frame_number = 100
#     h = 128
#     w = 128
#     # images = np.random.randint(0, 255, (frame_number, 512, 512, 3), dtype=np.uint8)
#     images = np.array(
#         [gsoup.generate_concentric_circles(h, w, n=5) / 255.0 for i in range(100)]
#     )
#     # im1 = gsoup.generate_voronoi_diagram(512, 512, 1000)
#     # im2 = gsoup.generate_voronoi_diagram(512, 512, 1000)
#     # im1s = np.tile(im1[None, ...], (10, 1, 1, 1))
#     # im2s = np.tile(im2[None, ...], (10, 1, 1, 1))
#     # images = np.vstack([im1s, im2s])
#     gsoup.save_video(images, Path("resource/lossless_video.avi"), lossy=False, fps=10)
#     gsoup.save_video(images, Path("resource/lossy_video.avi"), lossy=True, fps=10)
#     gsoup.trim_video(
#         Path("resource/lossless_video.avi"),
#         Path("resource/trimmed.avi"),
#         0,
#         100,
#         True,
#     )
#     reader = gsoup.VideoReader(
#         Path("resource/lossless_video.avi"), h=h, w=w, verbose=True
#     )
#     fps = gsoup.FPS()
#     reader_has_frames = False
#     for i, frame in enumerate(reader):
#         reader_has_frames = True
#         print("{}: {}, fps: {}".format(i, frame.shape, fps()))
#         assert np.all(frame == images[i])
#     assert reader_has_frames
#     video_frames = gsoup.load_video(Path("resource/lossless_video.avi"))
#     assert video_frames.shape == (frame_number, h, w, 3)
#     assert np.all(video_frames == images)
#     video_frames_reversed = gsoup.reverse_video(Path("resource/lossless_video.avi"))
#     assert (video_frames_reversed[-1] == video_frames[0]).all()
#     sliced_frames = gsoup.slice_from_video(
#         Path("resource/lossless_video.avi"),
#         every_n_frames=2,
#         start_frame=0,
#         end_frame=6,
#     )
#     assert (sliced_frames == video_frames[:7:2, :, :, :]).all()
#     gsoup.video_to_images(
#         Path("resource/lossless_video.avi"),
#         Path("resource/ffmpeg_reconstructed_images"),
#     )
#     gsoup.save_video(
#         Path("resource/ffmpeg_reconstructed_images"),
#         Path("resource/reconst_lossy.avi"),
#         fps=10,
#         lossy=True,
#     )
#     gsoup.save_video(
#         Path("resource/ffmpeg_reconstructed_images"),
#         Path("resource/reconst_lossless.avi"),
#         fps=10,
#         lossy=False,
#     )
#     discrete_images = gsoup.load_images(Path("resource/ffmpeg_reconstructed_images"))
#     assert discrete_images.shape == (frame_number, h, w, 3)
#     timestamps = gsoup.get_frame_timestamps(Path("resource/lossless_video.avi"))
#     assert timestamps[0] == 0


def test_procam():
    gray = gsoup.GrayCode()
    patterns = gray.encode((128, 128))
    mode = "ij"
    forward_map, fg = gray.decode(
        patterns, (128, 128), output_dir=Path("resource/pix2pix"), mode=mode, debug=True
    )
    backward_map = gsoup.compute_backward_map(
        (128, 128), forward_map, fg, output_dir=Path("resource/pix2pix"), debug=True
    )
    desired = gsoup.generate_lollipop_pattern(128, 128)
    warp_image = gsoup.warp_image(
        backward_map,
        desired,
        cam_wh=(forward_map.shape[1], forward_map.shape[0]),
        mode=mode,
        output_path=Path("resource/warp.png"),
    )
    assert warp_image.shape == (128, 128, 3)
    assert warp_image.dtype == np.uint8
    assert (
        np.mean(np.abs(desired - warp_image)) < 10
    )  # identity correspondence & warp should be very similar
    # calibration_dir = Path("resource/calibration")
    # calibration_dir.mkdir(exist_ok=True, parents=True)
    checkerboard = gsoup.generate_checkerboard(128, 128, 16)
    # T = gsoup.random_perspective()
    # T_opencv = T[:2, :]
    # img_transformed = cv2.warpPerspective(checkerboard, T, (128, 128))
    # captures = np.bitwise_and(patterns==255, checkerboard[None, ...]==1.0)
    # gsoup.save_images(captures, Path(calibration_dir, "0"))
    # gsoup.save_images(captures, Path(calibration_dir, "1"))
    gsoup.save_image(checkerboard, Path("resource/checkerboard.png"))
    #############
    # patterns = gray.encode((800, 800))
    patterns = gsoup.load_images(Path("tests/tests_resource/correspondence"))
    cam_wh = (patterns[0].shape[1], patterns[0].shape[0])
    proj_wh = (800, 800)
    forward_map, fg = gray.decode(
        patterns, (800, 800), output_dir="resource/forward", debug=True, mode=mode
    )
    backward_map = gsoup.compute_backward_map(
        (800, 800),
        forward_map,
        fg,
        mode=mode,
        output_dir="resource/backward_not_interp",
        debug=True,
        interpolate=False,
    )
    backward_map = gsoup.compute_backward_map(
        (800, 800),
        forward_map,
        fg,
        mode=mode,
        output_dir="resource/backward_interp",
        debug=True,
        interpolate=True,
    )
    desired = gsoup.generate_lollipop_pattern(800, 800)
    warp_image = gsoup.warp_image(
        backward_map,
        desired,
        cam_wh=cam_wh,
        mode=mode,
        output_path=Path("resource/debug/warp.png"),
    )
    blend_to_cv = np.array([[1.0, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]])
    cam_transform = np.array([[0, 0, 1, 1], [1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 0, 1.0]])
    cam_transform = cam_transform @ blend_to_cv
    ### gt ###
    # cam_int = np.array([[800, 0, 400.0], [0.0, 800, 400], [0, 0, 1]])
    # proj_int = np.array([[800, 0, 400.0], [0.0, 800, 400], [0, 0, 1]])
    # proj_transform = np.array([[-0.12403473, -0.23891242,  0.96308672,  0.8 ],
    #                             [ 0.99227786, -0.02986405,  0.12038584,  0.1],
    #                             [ 0.        ,  0.97058171,  0.24077168,  0.2],
    #                             [ 0.        ,  0.        ,  0.        ,  1. ]])
    # proj_transform = proj_transform @ blend_to_cv
    # cam_dist = None
    ### end gt ###
    ### calib ###
    result = gsoup.calibrate_procam(
        (800, 800),
        Path("tests/tests_resource/calibration"),
        chess_vert=15,
        chess_hori=15,
        chess_block_size=0.0185,
        output_dir="resource/calibration",
        projector_orientation="none",
        debug=True,
    )
    cam_int, cam_dist, proj_int, proj_dist, proj_transform = (
        result["cam_intrinsics"],
        result["cam_distortion"],
        result["proj_intrinsics"],
        result["proj_distortion"],
        result["proj_transform"],
    )
    proj_transform = cam_transform @ np.linalg.inv(proj_transform)  # p2w = c2w @ p2c
    ### end calib ###
    # calibration_static_view(cam_transform, proj_transform, (800, 800), (800, 800), cam_int, cam_dist, proj_int, forward_map, fg, mode)
    pc = gsoup.reconstruct_pointcloud(
        forward_map,
        fg,
        cam_transform,
        proj_transform,
        cam_int,
        cam_dist,
        proj_int,
        mode=mode,
    )
    gsoup.save_pointcloud(pc, "resource/points.ply")


def test_sphere_tracer():
    image_size = 512
    # device = "cuda:0"
    device = "cpu"
    w2v, v2c = gsoup.create_random_cameras_on_unit_sphere(
        5, 1.0, opengl=True, device=device
    )
    ray_origins, ray_directions = gsoup.generate_rays(
        w2v, v2c[0], image_size, image_size, device=device
    )
    sdf = gsoup.structures.sphere_sdf(0.25)
    images = []
    for o, d in zip(ray_origins, ray_directions):
        result = gsoup.render_sdf(sdf, o.view(-1, 3), d.view(-1, 3))
        images.append(result.view(image_size, image_size, 4))
    images = gsoup.to_np(torch.stack(images))
    images = gsoup.alpha_compose(images)
    gizmo_images = gsoup.draw_gizmo_on_image(
        images, gsoup.to_np(v2c @ w2v), opengl=True
    )
    gsoup.save_images(gizmo_images, Path("resource/sphere_trace"))


def test_geometry():
    faces_tri = np.array([[0, 1, 2], [2, 3, 0]])  # Triangle example
    faces_quad = np.array([[0, 1, 2, 3], [4, 5, 6, 7]])  # Quad example
    _, faces_quadcube = gsoup.structures.quad_cube()
    _, faces_tricube = gsoup.structures.cube()

    e, f2e = gsoup.faces2edges_naive(faces_tri)
    e, f2e = gsoup.faces2edges_naive(faces_tricube)
    e, f2e = gsoup.faces2edges_naive(faces_quad)
    e, f2e = gsoup.faces2edges_naive(faces_quadcube)


def test_rasterizer():
    # prep "canvas"
    width, height = 256, 256
    f = 800
    V3, F3 = gsoup.load_mesh("tests/tests_resource/cube.obj")  # (n, 3)
    V4, F4 = gsoup.structures.quad_cube()  # (n, 4)
    # Some random transformation
    V3 = V3 / 8
    V4 = V4 / 8
    rand_qvec = gsoup.random_qvec(1)
    rand_rot_mat = gsoup.qvec2mat(rand_qvec)
    rand_rot_trans = np.random.uniform(-0.1, 0.1, size=3)
    random_rigid = gsoup.compose_rt(
        rand_rot_mat, rand_rot_trans[None, ...], square=False
    )
    V3 = (random_rigid[0] @ gsoup.to_hom(V3).T).T
    V4 = (random_rigid[0] @ gsoup.to_hom(V4).T).T
    # Camera intrinsics matrix
    K = np.array([[f, 0, width / 2], [0, f, height / 2], [0, 0, 1]])
    K = K.astype(np.float32)
    # Camera extrinsics
    # Rt = np.eye(4)[:3, :]  # (identity for now)
    cam_loc = np.array([1.0, 0.0, 0.0])
    cam_at = np.array([0.0, 0.0, 0.0])
    cam_up = np.array([0.0, 0.0, 1.0])
    Rt = gsoup.look_at_np(cam_loc, cam_at, cam_up)  # cam -> world
    Rt = gsoup.invert_rigid(Rt)[0]  # world -> cam
    Rt = Rt.astype(np.float32)
    # Define colors per face
    colors3 = np.random.randint(0, 256, size=len(F3) * 3).reshape(len(F3), 3)
    colors4 = np.random.randint(0, 256, size=len(F4) * 3).reshape(len(F4), 3)
    colors1 = np.array([255, 255, 255])

    image = gsoup.render_mesh(V3, F3, K, Rt, colors3, wh=(width, height))
    gsoup.save_image(image, "resource/rast_tricube.png")

    image = gsoup.render_mesh(V4, F4, K, Rt, colors4, wh=(width, height))
    gsoup.save_image(image, "resource/rast_quadcube.png")

    image = gsoup.render_mesh(
        V4,
        F4,
        K,
        Rt,
        colors1,
        wireframe=True,
        wh=(width, height),
    )
    gsoup.save_image(image, "resource/rast_wireframe.png")

    image = gsoup.render_mesh(
        V4,
        F4,
        K,
        Rt,
        colors1,
        wireframe=True,
        wireframe_occlude=True,
        wh=(width, height),
    )
    gsoup.save_image(image, "resource/rast_wireframe_occlude.png")

    image = gsoup.render_mesh(
        V4,
        F4,
        K,
        Rt,
        colors4,
        wireframe=True,
        wh=(width, height),
    )
    gsoup.save_image(image, "resource/rast_wireframe_color.png")
    # test with providing image and depth_buffer
    image = np.zeros((height, width, 3), dtype=np.uint8)
    depth_buffer = np.full((height, width), np.inf, dtype=np.float32)
    gsoup.render_mesh(
        V4,
        F4,
        K,
        Rt,
        colors4,
        image=image,
        depth_buffer=depth_buffer,
    )
    gsoup.save_image(image, "resource/rast_provide_image.png")


def test_qem():
    v, f = gsoup.structures.cube()
    v_new, f_new = gsoup.qem(v, f, budget=4)
    assert f_new.shape[0] == 4
